# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
from copy import copy
from typing import Optional

import aiohttp
import requests
import storey
from storey.flow import _ConcurrentJobExecution

import mlrun
import mlrun.common.schemas
import mlrun.config
import mlrun.platforms
import mlrun.utils.async_http
from mlrun.errors import err_to_str
from mlrun.utils import dict_to_json, logger

from ..config import config
from .utils import (
    _extract_input_data,
    _update_result_body,
    event_id_key,
    event_path_key,
)

default_retries = 6
default_backoff_factor = 1


class RemoteStep(storey.SendToHttp):
    def __init__(
        self,
        url: str,
        subpath: Optional[str] = None,
        method: Optional[str] = None,
        headers: Optional[dict] = None,
        url_expression: Optional[str] = None,
        body_expression: Optional[str] = None,
        return_json: bool = True,
        input_path: Optional[str] = None,
        result_path: Optional[str] = None,
        max_in_flight=None,
        retries=None,
        backoff_factor=None,
        timeout=None,
        headers_expression: Optional[str] = None,
        **kwargs,
    ):
        """class for calling remote endpoints

        sync and async graph step implementation for request/resp to remote service (class shortcut = "$remote")
        url can be an http(s) url (e.g. `https://myservice/path`) or an mlrun function uri ([project/]name).
        alternatively the url_expression can be specified to build the url from the event (e.g. "event['url']").

        example pipeline::

            flow = function.set_topology("flow", engine="async")
            flow.to(name="step1", handler="func1")
                .to(RemoteStep(name="remote_echo", url="https://myservice/path", method="POST"))
                .to(name="laststep", handler="func2").respond()


        :param url:     http(s) url or function [project/]name to call
        :param subpath: path (which follows the url), use `$path` to use the event.path
        :param method:  The HTTP method to use for the request (e.g., "GET", "POST", "PUT", "DELETE").
                        If not provided, the step will try to use `event.method` at runtime, and if that
                        is also missing, it defaults to `"POST"`.
        :param headers: dictionary with http header values
        :param url_expression: an expression for getting the url from the event, e.g. "event['url']"
        :param body_expression: an expression for getting the request body from the event, e.g. "event['data']"
        :param return_json: indicate the returned value is json, and convert it to a py object
        :param input_path:  when specified selects the key/path in the event to use as body
                            this require that the event body will behave like a dict, example:
                            event: {"data": {"a": 5, "b": 7}}, input_path="data.b" means request body will be 7
        :param result_path: selects the key/path in the event to write the results to
                            this require that the event body will behave like a dict, example:
                            event: {"x": 5} , result_path="resp" means the returned response will be written
                            to event["y"] resulting in {"x": 5, "resp": <result>}
        :param retries:     number of retries (in exponential backoff)
        :param backoff_factor: A backoff factor in seconds to apply between attempts after the second try
        :param timeout:     How long to wait for the server to send data before giving up, float in seconds
        :param headers_expression: an expression for getting the request headers from the event, e.g. "event['headers']"
        """
        # init retry args for storey
        retries = default_retries if retries is None else retries
        super().__init__(
            None,
            None,
            input_path=input_path,
            result_path=result_path,
            max_in_flight=max_in_flight,
            retries=retries,
            backoff_factor=backoff_factor,
            **kwargs,
        )
        self.url = url
        self.url_expression = url_expression
        self.body_expression = body_expression
        self.headers_expression = headers_expression
        self.headers = headers
        self.method = method
        self.return_json = return_json
        self.subpath = subpath

        self.timeout = timeout

        self._append_event_path = False
        self._endpoint = ""
        self._session = None
        self._url_function_handler = None
        self._body_function_handler = None
        self._headers_function_handler = None

    def post_init(self, mode="sync", **kwargs):
        self._endpoint = self.url
        if self.url and self.context:
            self._endpoint = self.context.get_remote_endpoint(self.url).strip("/")
        if self.body_expression:
            # init lambda function for calculating url from event
            self._body_function_handler = eval(
                "lambda event: " + self.body_expression, {"context": self.context}, {}
            )
        if self.url_expression:
            # init lambda function for calculating url from event
            self._url_function_handler = eval(
                "lambda event: " + self.url_expression,
                {"endpoint": self._endpoint, "context": self.context},
                {},
            )
        if self.headers_expression:
            self._headers_function_handler = eval(
                "lambda event: " + self.headers_expression,
                {"context": self.context},
                {},
            )
        elif self.subpath:
            self._append_event_path = self.subpath == "$path"
            if not self._append_event_path:
                self._endpoint = self._endpoint + "/" + self.subpath.lstrip("/")

    async def _process_event(self, event):
        # async implementation (with storey)
        body = self._get_event_or_body(event)
        method, url, headers, body, kwargs = self._generate_request(event, body)
        kwargs = kwargs or {}
        if self.timeout:
            kwargs["timeout"] = aiohttp.ClientTimeout(total=self.timeout)
        try:
            resp = await self._client_session.request(
                method, url, headers=headers, data=body, ssl=False, **kwargs
            )
            if resp.status >= 500:
                text = await resp.text()
                raise RuntimeError(f"bad http response {resp.status}: {text}")
            return resp
        except asyncio.TimeoutError as exc:
            logger.error(f"http request to {url} timed out in RemoteStep {self.name}")
            raise exc

    async def _handle_completed(self, event, response):
        response_body = await response.read()
        if response.status >= 400:
            raise ValueError(
                f"For event {event}, RemoteStep {self.name} got an unexpected response "
                f"status {response.status}: {response_body}"
            )

        body = self._get_data(response_body, response.headers)

        new_event = self._user_fn_output_to_event(event, body)
        await self._do_downstream(new_event)

    def do_event(self, event):
        # sync implementation (without storey)
        if not self._session:
            self._session = mlrun.utils.HTTPSessionWithRetry(
                self.retries,
                self.backoff_factor or mlrun.mlconf.http_retry_defaults.backoff_factor,
                retry_on_exception=False,
                retry_on_status=self.retries > 0,
                retry_on_post=True,
            )

        body = _extract_input_data(self._input_path, event.body)
        method, url, headers, body, kwargs = self._generate_request(event, body)
        try:
            resp = self._session.request(
                method,
                url,
                verify=mlrun.mlconf.httpdb.http.verify,
                headers=headers,
                data=body,
                timeout=self.timeout,
                **kwargs,
            )
        except requests.exceptions.ReadTimeout as err:
            raise requests.exceptions.ReadTimeout(
                f"http request to {url} timed out in RemoteStep {self.name}, {err_to_str(err)}"
            )
        except OSError as err:
            raise OSError(f"cannot invoke url: {url}, {err_to_str(err)}")
        if not resp.ok:
            raise RuntimeError(f"bad http response {resp.status_code}: {resp.text}")

        result = self._get_data(resp.content, resp.headers)
        event.body = _update_result_body(self._result_path, event.body, result)
        return event

    def _generate_request(self, event, body):
        method = self.method or event.method or "POST"
        if self._headers_function_handler:
            headers = self._headers_function_handler(body)
        else:
            headers = copy(self.headers) or {}

        if self._url_function_handler:
            url = self._url_function_handler(body)
        else:
            url = self._endpoint
            striped_path = event.path.lstrip("/")
            if self._append_event_path:
                url = url + "/" + striped_path
            if striped_path:
                headers[event_path_key] = event.path
        if event.id:
            headers[event_id_key] = event.id
        if method == "GET":
            body = None
        elif body is not None and not isinstance(body, (str, bytes)):
            if self._body_function_handler:
                body = self._body_function_handler(body)
            body = json.dumps(body)
            headers["Content-Type"] = "application/json"

        return method, url, headers, body, {}

    def _get_data(self, data, headers):
        if (
            self.return_json
            or headers.get("content-type", "").lower() == "application/json"
        ) and isinstance(data, (str, bytes)):
            data = json.loads(data)
        return data


class BatchHttpRequests(_ConcurrentJobExecution):
    def __init__(
        self,
        url: Optional[str] = None,
        subpath: Optional[str] = None,
        method: Optional[str] = None,
        headers: Optional[dict] = None,
        url_expression: Optional[str] = None,
        body_expression: Optional[str] = None,
        return_json: bool = True,
        input_path: Optional[str] = None,
        result_path: Optional[str] = None,
        retries=None,
        backoff_factor=None,
        timeout=None,
        **kwargs,
    ):
        """class for calling remote endpoints in parallel

        sync and async graph step implementation for request/resp to remote service (class shortcut = "$remote")
        url can be an http(s) url (e.g. `https://myservice/path`) or an mlrun function uri ([project/]name).
        alternatively the url_expression can be specified to build the url from the event (e.g. "event['url']").

        example pipeline::

            function = mlrun.new_function("myfunc", kind="serving")
            flow = function.set_topology("flow", engine="async")
            flow.to(
                BatchHttpRequests(
                    url_expression="event['url']",
                    body_expression="event['data']",
                    method="POST",
                    input_path="req",
                    result_path="resp",
                )
            ).respond()

            server = function.to_mock_server()
            # request contains a list of elements, each with url and data
            request = [{"url": f"{base_url}/{i}", "data": i} for i in range(2)]
            resp = server.test(body={"req": request})


        :param url:     http(s) url or function [project/]name to call
        :param subpath: path (which follows the url)
        :param method:  HTTP method (GET, POST, ..), default to POST
        :param headers: dictionary with http header values
        :param url_expression: an expression for getting the url from the event, e.g. "event['url']"
        :param body_expression: an expression for getting the request body from the event, e.g. "event['data']"
        :param return_json: indicate the returned value is json, and convert it to a py object
        :param input_path:  when specified selects the key/path in the event to use as body
                            this require that the event body will behave like a dict, example:
                            event: {"data": {"a": 5, "b": 7}}, input_path="data.b" means request body will be 7
        :param result_path: selects the key/path in the event to write the results to
                            this require that the event body will behave like a dict, example:
                            event: {"x": 5} , result_path="resp" means the returned response will be written
                            to event["y"] resulting in {"x": 5, "resp": <result>}
        :param retries:     number of retries (in exponential backoff)
        :param backoff_factor: A backoff factor in seconds to apply between attempts after the second try
        :param timeout:     How long to wait for the server to send data before giving up, float in seconds
        """
        if url and url_expression:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "cannot set both url and url_expression"
            )
        self.url = url
        self.url_expression = url_expression
        self.body_expression = body_expression
        self.headers = headers
        self.method = method
        self.return_json = return_json
        self.subpath = subpath
        super().__init__(input_path=input_path, result_path=result_path, **kwargs)

        self.timeout = timeout
        self.retries = retries
        self.backoff_factor = backoff_factor
        self._append_event_path = False
        self._endpoint = ""
        self._session = None
        self._url_function_handler = None
        self._body_function_handler = None
        self._request_args = {}

    def _init(self):
        super()._init()
        self._client_session = None

    async def _lazy_init(self):
        connector = aiohttp.TCPConnector()
        self._client_session = aiohttp.ClientSession(connector=connector)

    async def _cleanup(self):
        await self._client_session.close()

    def post_init(self, mode="sync", **kwargs):
        self._endpoint = self.url
        if self.url and self.context:
            self._endpoint = self.context.get_remote_endpoint(self.url).strip("/")
        if self.body_expression:
            # init lambda function for calculating url from event
            self._body_function_handler = eval(
                "lambda event: " + self.body_expression, {}, {}
            )
        if self.url_expression:
            # init lambda function for calculating url from event
            self._url_function_handler = eval(
                "lambda event: " + self.url_expression, {}, {"endpoint": self._endpoint}
            )
        elif self.subpath:
            self._endpoint = self._endpoint + "/" + self.subpath.lstrip("/")
        if self.timeout:
            self._request_args["timeout"] = aiohttp.ClientTimeout(total=self.timeout)

    async def _process_event(self, event):
        # async implementation (with storey)
        method = self.method or event.method or "POST"
        headers = self.headers or {}
        input_list = self._get_event_or_body(event)
        is_get = method == "GET"
        is_json = False

        body_list = []
        url_list = []
        for body in input_list:
            if self._url_function_handler:
                url_list.append(self._url_function_handler(body))
            else:
                url_list.append(self._endpoint)

            if is_get:
                body = None
            elif body is not None and not isinstance(body, (str, bytes)):
                if self._body_function_handler:
                    body = self._body_function_handler(body)
                body = json.dumps(body)
                is_json = True
            body_list.append(body)

        if is_json:
            headers["Content-Type"] = "application/json"

        responses = []
        for url, body in zip(url_list, body_list):
            responses.append(
                asyncio.ensure_future(
                    self._submit_with_retries(method, url, headers, body)
                )
            )
        return await asyncio.gather(*responses)

    async def _process_event_with_retries(self, event):
        return await self._process_event(event)

    async def _submit(self, method, url, headers, body):
        async with self._client_session.request(
            method, url, headers=headers, data=body, ssl=False, **self._request_args
        ) as future:
            if future.status >= 500:
                text = await future.text()
                raise RuntimeError(f"bad http response {future.status}: {text}")
            return await future.read(), future.headers

    async def _submit_with_retries(self, method, url, headers, body):
        times_attempted = 0
        max_attempts = (self.retries or default_retries) + 1
        while True:
            try:
                return await self._submit(method, url, headers, body)
            except Exception as ex:
                times_attempted += 1
                attempts_left = max_attempts - times_attempted
                if self.logger:
                    self.logger.warn(
                        f"{self.name} failed to process event ({attempts_left} retries left): {ex}"
                    )
                if attempts_left <= 0:
                    raise ex
                backoff_factor = (
                    default_backoff_factor
                    if self.backoff_factor is None
                    else self.backoff_factor
                )
                backoff_value = (backoff_factor) * (2 ** (times_attempted - 1))
                backoff_value = min(self._BACKOFF_MAX, backoff_value)
                if backoff_value >= 0:
                    await asyncio.sleep(backoff_value)

    async def _handle_completed(self, event, response):
        data = []
        for body, headers in response:
            data.append(self._get_data(body, headers))

        new_event = self._user_fn_output_to_event(event, data)
        await self._do_downstream(new_event)

    def _get_data(self, data, headers):
        if (
            self.return_json
            or headers.get("content-type", "").lower() == "application/json"
        ) and isinstance(data, (str, bytes)):
            data = json.loads(data)
        return data


class MLRunAPIRemoteStep(RemoteStep):
    def __init__(
        self, method: str, path: str, fill_placeholders: Optional[bool] = None, **kwargs
    ):
        """
        Graph step implementation for calling MLRun API endpoints

        :param method:  The HTTP method to use for the request (e.g., "GET", "POST", "PUT", "DELETE").
                        If not provided, the step will try to use `event.method` at runtime, and if that
                        is also missing, it defaults to `"POST"`.
        :param path:    API path (e.g. /api/projects)
        :param fill_placeholders: if True, fill placeholders in the path using event fields (default to False)
        :param kwargs:  other arguments passed to RemoteStep
        """
        super().__init__(url="", method=method, **kwargs)
        self.rundb = None
        self.path = path
        self.fill_placeholders = fill_placeholders

    def _generate_request(self, event, body):
        method = self.method or event.method or "POST"
        kw = {
            key: value
            for key, value in (
                ("params", body.get("params")),
                ("json", body.get("json")),
            )
            if value is not None
        }

        headers = self.headers or {}
        headers.update(body.get("headers", {}))

        if self.rundb.user:
            kw["auth"] = (self.rundb.user, self.rundb.password)
        elif self.rundb.token_provider:
            token = self.rundb.token_provider.get_token()
            if token:
                # Iguazio auth doesn't support passing token through bearer, so use cookie instead
                if self.rundb.token_provider.is_iguazio_session():
                    session_cookie = f'session=j:{{"sid": "{token}"}}'
                    headers["cookie"] = session_cookie
                else:
                    if "Authorization" not in kw.setdefault("headers", {}):
                        headers.update({"Authorization": "Bearer " + token})

        if mlrun.common.schemas.HeaderNames.client_version not in headers:
            headers.update(
                {
                    mlrun.common.schemas.HeaderNames.client_version: self.rundb.client_version,
                    mlrun.common.schemas.HeaderNames.python_version: self.rundb.python_version,
                    "User-Agent": f"{requests.utils.default_user_agent()} mlrun/{config.version}",
                }
            )

        url = self.url.format(**body) if self.fill_placeholders else self.url
        headers["Content-Type"] = "application/json"
        return method, url, headers, dict_to_json(body), kw

    def post_init(self, mode="sync", **kwargs):
        super().post_init(mode=mode, **kwargs)
        self.fill_placeholders = self.fill_placeholders or False
        self.rundb = mlrun.get_run_db()
        self.url = self.rundb.get_base_api_url(self.path)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import operator
import os
import warnings
from base64 import b64encode
from copy import copy
from datetime import datetime
from typing import Any, Literal, Optional, Union

import pandas as pd
import semver
import v3io
import v3io.dataplane
from nuclio import KafkaTrigger
from nuclio.config import split_path

import mlrun
from mlrun.config import config
from mlrun.datastore.snowflake_utils import get_snowflake_spark_options
from mlrun.datastore.utils import transform_list_filters_to_tuple
from mlrun.secrets import SecretsStore
from mlrun.utils import logger

from ..common.schemas.function import Function
from ..model import DataSource
from ..platforms.iguazio import parse_path
from ..utils import get_class, is_explicit_ack_supported
from .datastore_profile import datastore_profile_read
from .spark_utils import spark_session_update_hadoop_options
from .utils import (
    _generate_sql_query_with_time_filter,
    filter_df_start_end_time,
    select_columns_from_df,
)


def load_spark_dataframe_with_options(session, spark_options, format=None):
    non_hadoop_spark_options = spark_session_update_hadoop_options(
        session, spark_options
    )
    if format:
        df = session.read.format(format).load(**non_hadoop_spark_options)
    else:
        df = session.read.load(**non_hadoop_spark_options)
    return df


def get_source_from_dict(source):
    kind = source.get("kind", "")
    if not kind:
        return None
    return source_kind_to_driver[kind].from_dict(source)


def get_source_step(source, key_fields=None, time_field=None, context=None):
    """initialize the source driver"""
    if hasattr(source, "to_csv"):
        source = DataFrameSource(source, context=context)
    if not key_fields and not source.key_field:
        raise mlrun.errors.MLRunInvalidArgumentError("key column is not defined")
    return source.to_step(key_fields, time_field, context)


class BaseSourceDriver(DataSource):
    support_spark = False
    support_storey = False

    def to_step(self, key_field=None, time_field=None, context=None):
        import storey

        if not self.support_storey:
            raise mlrun.errors.MLRunRuntimeError(
                f"{type(self).__name__} does not support storey engine"
            )

        explicit_ack = (
            is_explicit_ack_supported(context)
            and mlrun.mlconf.is_explicit_ack_enabled()
        )
        return storey.SyncEmitSource(
            context=context,
            explicit_ack=explicit_ack,
        )

    def get_table_object(self):
        """get storey Table object"""
        return None

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        """return the source data as dataframe"""
        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        return mlrun.store_manager.object(url=self.path).as_df(
            columns=columns,
            df_module=df_module,
            start_time=start_time or self.start_time,
            end_time=end_time or self.end_time,
            time_column=time_field or self.time_field,
        )

    def to_spark_df(self, session, named_view=False, time_field=None, columns=None):
        if self.support_spark:
            spark_options = self.get_spark_options()
            spark_format = spark_options.pop("format", None)
            df = load_spark_dataframe_with_options(
                session, spark_options, format=spark_format
            )
            if named_view:
                df.createOrReplaceTempView(self.name)
            return self._filter_spark_df(df, time_field, columns)
        raise NotImplementedError(
            f"Conversion of a source of type '{type(self).__name__}' "
            "to a Spark dataframe is not possible, as this operation is not supported"
        )

    def _filter_spark_df(self, df, time_field=None, columns=None):
        if not (columns or time_field):
            return df

        from pyspark.sql.functions import col

        if time_field:
            if self.start_time:
                df = df.filter(col(time_field) > self.start_time)
            if self.end_time:
                df = df.filter(col(time_field) <= self.end_time)

        if columns:
            df = df.select([col(name) for name in columns])
        return df

    def get_spark_options(self):
        # options used in spark.read.load(**options)
        raise NotImplementedError()

    def is_iterator(self):
        return False


class CSVSource(BaseSourceDriver):
    """
    Reads CSV file as input source for a flow.

    :parameter name: name of the source
    :parameter path: path to CSV file
    :parameter key_field: the CSV field to be used as the key for events. May be an int (field index) or string
        (field name) if with_header is True. Defaults to None (no key). Can be a list of keys.
    :parameter schedule: string to configure scheduling of the ingestion job.
    :parameter attributes: additional parameters to pass to storey. For example:
        attributes={"timestamp_format": '%Y%m%d%H'}
    :parameter parse_dates: Optional. List of columns (names or integers) that will be
        attempted to parse as date column.
    """

    kind = "csv"
    support_storey = True
    support_spark = True

    def __init__(
        self,
        name: str = "",
        path: Optional[str] = None,
        attributes: Optional[dict[str, object]] = None,
        key_field: Optional[str] = None,
        schedule: Optional[str] = None,
        parse_dates: Union[None, int, str, list[int], list[str]] = None,
        **kwargs,
    ):
        super().__init__(name, path, attributes, key_field, schedule=schedule, **kwargs)
        if parse_dates and not isinstance(parse_dates, list):
            parse_dates = [parse_dates]
        self._parse_dates = parse_dates

    def to_step(self, key_field=None, time_field=None, context=None):
        import storey

        attributes = self.attributes or {}
        if context:
            attributes["context"] = context

        parse_dates = self._parse_dates or []
        if time_field and time_field not in parse_dates:
            parse_dates.append(time_field)

        data_item = mlrun.store_manager.object(self.path)
        store, path, url = mlrun.store_manager.get_or_create_store(self.path)

        return storey.CSVSource(
            paths=url,  # unlike self.path, it already has store:// replaced
            build_dict=True,
            key_field=self.key_field or key_field,
            storage_options=data_item.store.get_storage_options(),
            parse_dates=parse_dates,
            **attributes,
        )

    def get_spark_options(self):
        store, path, _ = mlrun.store_manager.get_or_create_store(self.path)
        spark_options = store.get_spark_options(store.spark_url + path)
        spark_options.update(
            {
                "path": store.spark_url + path,
                "format": "csv",
                "header": "true",
                "inferSchema": "true",
            }
        )
        return spark_options

    def to_spark_df(self, session, named_view=False, time_field=None, columns=None):
        import pyspark.sql.functions as funcs

        df = load_spark_dataframe_with_options(session, self.get_spark_options())

        parse_dates = self._parse_dates or []
        if time_field and time_field not in parse_dates:
            parse_dates.append(time_field)

        for col_name, col_type in df.dtypes:
            if parse_dates and col_name in parse_dates:
                df = df.withColumn(col_name, funcs.col(col_name).cast("timestamp"))
        if named_view:
            df.createOrReplaceTempView(self.name)
        return self._filter_spark_df(df, time_field, columns)

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        reader_args = self.attributes.get("reader_args", {})
        return mlrun.store_manager.object(url=self.path).as_df(
            columns=columns,
            df_module=df_module,
            format="csv",
            start_time=start_time or self.start_time,
            end_time=end_time or self.end_time,
            time_column=time_field or self.time_field,
            parse_dates=self._parse_dates,
            chunksize=self.attributes.get("chunksize"),
            **reader_args,
        )

    def is_iterator(self):
        return bool(self.attributes.get("chunksize"))


class ParquetSource(BaseSourceDriver):
    """
    Reads Parquet file/dir as input source for a flow.

    :parameter name: name of the source
    :parameter path: path to Parquet file or directory
    :parameter key_field: the column to be used as the key for events. Can be a list of keys.
    :parameter time_field: Optional. Feature set's timestamp_key will be used if None. The results will be filtered
         by this column and start_filter & end_filter.
    :parameter start_filter: datetime. If not None, the results will be filtered by partitions and
         'filter_column' > start_filter. Default is None
    :parameter end_filter: datetime. If not None, the results will be filtered by partitions
         'filter_column' <= end_filter. Default is None
    :parameter schedule: string to configure scheduling of the ingestion job. For example `'*/30 * * * *'` will
         cause the job to run every 30 minutes
    :parameter start_time: filters out data before this time
    :parameter end_time: filters out data after this time
    :parameter attributes: additional parameters to pass to storey.
    :param additional_filters: List of additional_filter conditions as tuples.
                               Each tuple should be in the format (column_name, operator, value).
                               Supported operators: "=", ">=", "<=", ">", "<".
                               Example: [("Product", "=", "Computer")]
                               For all supported filters, please see:
                               https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
    """

    kind = "parquet"
    support_storey = True
    support_spark = True

    def __init__(
        self,
        name: str = "",
        path: Optional[str] = None,
        attributes: Optional[dict[str, object]] = None,
        key_field: Optional[str] = None,
        time_field: Optional[str] = None,
        schedule: Optional[str] = None,
        start_time: Optional[Union[datetime, str]] = None,
        end_time: Optional[Union[datetime, str]] = None,
        additional_filters: Optional[list[Union[tuple, list]]] = None,
    ):
        if additional_filters:
            attributes = copy(attributes) or {}
            additional_filters = transform_list_filters_to_tuple(additional_filters)
            attributes["additional_filters"] = additional_filters

        super().__init__(
            name,
            path,
            attributes,
            key_field,
            time_field,
            schedule,
            start_time,
            end_time,
        )

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        self._start_time = self._convert_to_datetime(start_time)

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        self._end_time = self._convert_to_datetime(end_time)

    @property
    def additional_filters(self):
        return self.attributes.get("additional_filters")

    @staticmethod
    def _convert_to_datetime(time):
        if time and isinstance(time, str):
            if time.endswith("Z"):
                return datetime.fromisoformat(time.replace("Z", "+00:00"))
            return datetime.fromisoformat(time)
        else:
            return time

    def to_step(
        self,
        key_field=None,
        time_field=None,
        start_time=None,
        end_time=None,
        context=None,
        additional_filters=None,
    ):
        import storey

        attributes = copy(self.attributes)
        attributes.pop("additional_filters", None)
        if context:
            attributes["context"] = context
        additional_filters = transform_list_filters_to_tuple(additional_filters)
        data_item = mlrun.store_manager.object(self.path)
        store, path, url = mlrun.store_manager.get_or_create_store(self.path)
        return storey.ParquetSource(
            paths=url,  # unlike self.path, it already has store:// replaced
            key_field=self.key_field or key_field,
            storage_options=data_item.store.get_storage_options(),
            end_filter=self.end_time,
            start_filter=self.start_time,
            filter_column=self.time_field or time_field,
            additional_filters=self.additional_filters or additional_filters,
            **attributes,
        )

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: Optional[dict] = None
    ):
        new_obj = super().from_dict(
            struct=struct, fields=fields, deprecated_fields=deprecated_fields
        )
        new_obj.attributes["additional_filters"] = transform_list_filters_to_tuple(
            new_obj.additional_filters
        )
        return new_obj

    def get_spark_options(self):
        store, path, _ = mlrun.store_manager.get_or_create_store(self.path)
        spark_options = store.get_spark_options(store.spark_url + path)
        spark_options.update(
            {
                "path": store.spark_url + path,
                "format": "parquet",
            }
        )
        return spark_options

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        reader_args = self.attributes.get("reader_args", {})
        additional_filters = transform_list_filters_to_tuple(additional_filters)
        return mlrun.store_manager.object(url=self.path).as_df(
            columns=columns,
            df_module=df_module,
            start_time=start_time or self.start_time,
            end_time=end_time or self.end_time,
            time_column=time_field or self.time_field,
            format="parquet",
            additional_filters=additional_filters or self.additional_filters,
            **reader_args,
        )

    def _build_spark_additional_filters(self, column_types: dict):
        if not self.additional_filters:
            return None
        from pyspark.sql.functions import col, isnan, lit

        operators = {
            "==": operator.eq,
            "=": operator.eq,
            ">": operator.gt,
            "<": operator.lt,
            ">=": operator.ge,
            "<=": operator.le,
            "!=": operator.ne,
        }

        spark_filter = None
        new_filter = lit(True)
        for filter_tuple in self.additional_filters:
            if not filter_tuple:
                continue
            col_name, op, value = filter_tuple
            if op.lower() in ("in", "not in") and isinstance(value, (list, tuple, set)):
                none_exists = False
                value = list(value)
                for sub_value in value:
                    if sub_value is None:
                        value.remove(sub_value)
                        none_exists = True
                if none_exists:
                    filter_nan = column_types[col_name] not in ("timestamp", "date")
                    if value:
                        if op.lower() == "in":
                            new_filter = (
                                col(col_name).isin(value) | col(col_name).isNull()
                            )
                            if filter_nan:
                                new_filter = new_filter | isnan(col(col_name))

                        else:
                            new_filter = (
                                ~col(col_name).isin(value) & ~col(col_name).isNull()
                            )
                            if filter_nan:
                                new_filter = new_filter & ~isnan(col(col_name))
                    else:
                        if op.lower() == "in":
                            new_filter = col(col_name).isNull()
                            if filter_nan:
                                new_filter = new_filter | isnan(col(col_name))
                        else:
                            new_filter = ~col(col_name).isNull()
                            if filter_nan:
                                new_filter = new_filter & ~isnan(col(col_name))
                else:
                    if op.lower() == "in":
                        new_filter = col(col_name).isin(value)
                    elif op.lower() == "not in":
                        new_filter = ~col(col_name).isin(value)
            elif op in operators:
                new_filter = operators[op](col(col_name), value)
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"unsupported filter operator: {op}"
                )
            if spark_filter is not None:
                spark_filter = spark_filter & new_filter
            else:
                spark_filter = new_filter
        return spark_filter

    def _filter_spark_df(self, df, time_field=None, columns=None):
        spark_additional_filters = self._build_spark_additional_filters(
            column_types=dict(df.dtypes)
        )
        if spark_additional_filters is not None:
            df = df.filter(spark_additional_filters)
        return super()._filter_spark_df(df=df, time_field=time_field, columns=columns)


class BigQuerySource(BaseSourceDriver):
    """
    Reads Google BigQuery query results as input source for a flow.

    For authentication, set the GCP_CREDENTIALS project secret to the credentials json string.

    example::

         # set the credentials
         project.set_secrets({"GCP_CREDENTIALS": gcp_credentials_json})

         # use sql query
         query_string = "SELECT * FROM `the-psf.pypi.downloads20210328` LIMIT 5000"
         source = BigQuerySource(
             "bq1",
             query=query_string,
             gcp_project="my_project",
             materialization_dataset="dataviews",
         )

         # read a table
         source = BigQuerySource(
             "bq2", table="the-psf.pypi.downloads20210328", gcp_project="my_project"
         )


    :parameter name: source name
    :parameter table: table name/path, cannot be used together with query
    :parameter query: sql query string
    :parameter materialization_dataset: for query with spark, The target dataset for the materialized view.
                                        This dataset should be in same location as the view or the queried tables.
                                        must be set to a dataset where the GCP user has table creation permission
    :parameter chunksize: number of rows per chunk (default large single chunk)
    :parameter key_field: the column to be used as the key for events. Can be a list of keys.
    :parameter time_field: the column to be used for time filtering. Defaults to the feature set's timestamp_key.
    :parameter schedule: string to configure scheduling of the ingestion job. For example `'*/30 * * * *'` will
         cause the job to run every 30 minutes
    :parameter start_time: filters out data before this time
    :parameter end_time: filters out data after this time
    :parameter gcp_project: google cloud project name
    :parameter spark_options: additional spark read options
    """

    kind = "bigquery"
    support_storey = False
    support_spark = True

    def __init__(
        self,
        name: str = "",
        table: Optional[str] = None,
        max_results_for_table: Optional[int] = None,
        query: Optional[str] = None,
        materialization_dataset: Optional[str] = None,
        chunksize: Optional[int] = None,
        key_field: Optional[str] = None,
        time_field: Optional[str] = None,
        schedule: Optional[str] = None,
        start_time=None,
        end_time=None,
        gcp_project: Optional[str] = None,
        spark_options: Optional[dict] = None,
        **kwargs,
    ):
        if query and table:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "cannot specify both table and query args"
            )
        # Otherwise, the client library does not fully respect the limit
        if (
            max_results_for_table
            and chunksize
            and max_results_for_table % chunksize != 0
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "max_results_for_table must be a multiple of chunksize"
            )
        attrs = {
            "query": query,
            "table": table,
            "max_results": max_results_for_table,
            "chunksize": chunksize,
            "gcp_project": gcp_project,
            "spark_options": spark_options,
            "materialization_dataset": materialization_dataset,
        }
        attrs = {key: value for key, value in attrs.items() if value is not None}
        super().__init__(
            name,
            attributes=attrs,
            key_field=key_field,
            time_field=time_field,
            schedule=schedule,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )

    def _get_credentials_string(self):
        gcp_project = self.attributes.get("gcp_project", None)
        key = "GCP_CREDENTIALS"
        gcp_cred_string = os.getenv(key) or os.getenv(
            SecretsStore.k8s_env_variable_name_for_secret(key)
        )
        return gcp_cred_string, gcp_project

    def _get_credentials(self):
        from google.oauth2 import service_account

        gcp_cred_string, gcp_project = self._get_credentials_string()
        if gcp_cred_string and not os.getenv("GOOGLE_APPLICATION_CREDENTIALS"):
            gcp_cred_dict = json.loads(gcp_cred_string, strict=False)
            credentials = service_account.Credentials.from_service_account_info(
                gcp_cred_dict
            )
            return credentials, gcp_project or gcp_cred_dict["project_id"]
        return None, gcp_project

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        from google.cloud import bigquery
        from google.cloud.bigquery_storage_v1 import BigQueryReadClient

        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )

        def schema_to_dtypes(schema):
            from mlrun.data_types.data_types import gbq_to_pandas_dtype

            dtypes = {}
            for field in schema:
                dtypes[field.name] = gbq_to_pandas_dtype(field.field_type)
            return dtypes

        credentials, gcp_project = self._get_credentials()
        bqclient = bigquery.Client(project=gcp_project, credentials=credentials)

        query = self.attributes.get("query")
        table = self.attributes.get("table")
        chunksize = self.attributes.get("chunksize")
        if query:
            query_job = bqclient.query(query)

            rows_iterator = query_job.result(page_size=chunksize)
        elif table:
            table = self.attributes.get("table")
            max_results = self.attributes.get("max_results")

            rows_iterator = bqclient.list_rows(
                table, page_size=chunksize, max_results=max_results
            )
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "table or query args must be specified"
            )

        dtypes = schema_to_dtypes(rows_iterator.schema)
        if chunksize:
            # passing bqstorage_client greatly improves performance
            df = rows_iterator.to_dataframe_iterable(
                bqstorage_client=BigQueryReadClient(), dtypes=dtypes
            )
        else:
            df = rows_iterator.to_dataframe(dtypes=dtypes)

        return select_columns_from_df(
            filter_df_start_end_time(
                df,
                time_column=time_field or self.time_field,
                start_time=start_time or self.start_time,
                end_time=end_time or self.end_time,
            ),
            columns=columns,
        )

    def is_iterator(self):
        return bool(self.attributes.get("chunksize"))

    def to_spark_df(self, session, named_view=False, time_field=None, columns=None):
        options = copy(self.attributes.get("spark_options", {}))
        credentials, gcp_project = self._get_credentials_string()
        if credentials:
            options["credentials"] = b64encode(credentials.encode("utf-8")).decode(
                "utf-8"
            )
        if gcp_project:
            options["parentProject"] = gcp_project
        query = self.attributes.get("query")
        table = self.attributes.get("table")
        materialization_dataset = self.attributes.get("materialization_dataset")
        if not query and not table:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "table or query args must be specified"
            )
        if query and not materialization_dataset:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "materialization_dataset must be specified when running a query"
            )
        if query:
            options["viewsEnabled"] = True
            options["materializationDataset"] = materialization_dataset
            options["query"] = query
        elif table:
            options["path"] = table

        df = load_spark_dataframe_with_options(session, options, "bigquery")
        if named_view:
            df.createOrReplaceTempView(self.name)
        return self._filter_spark_df(df, time_field, columns)


class SnowflakeSource(BaseSourceDriver):
    """
    Reads Snowflake query results as input source for a flow.

    The Snowflake cluster's password must be provided using the SNOWFLAKE_PASSWORD environment variable or secret.
    See https://docs.mlrun.org/en/latest/store/datastore.html#storage-credentials-and-parameters for how to set secrets.

    example::

         source = SnowflakeSource(
            "sf",
            query="..",
            url="...",
            user="...",
            database="...",
            db_schema="...",
            warehouse="...",
        )

    :parameter name: source name
    :parameter key_field: the column to be used as the key for events. Can be a list of keys.
    :parameter time_field: the column to be used for time filtering. Defaults to the feature set's timestamp_key.
    :parameter schedule: string to configure scheduling of the ingestion job. For example `'*/30 * * * *'` will
         cause the job to run every 30 minutes
    :parameter start_time: filters out data before this time
    :parameter end_time: filters out data after this time
    :parameter query: sql query string
    :parameter url: URL of the snowflake cluster
    :parameter user: snowflake user
    :parameter database: snowflake database
    :parameter db_schema: snowflake schema
    :parameter warehouse: snowflake warehouse
    """

    kind = "snowflake"
    support_spark = True
    support_storey = False

    def __init__(
        self,
        name: str = "",
        key_field: Optional[str] = None,
        attributes: Optional[dict[str, object]] = None,
        time_field: Optional[str] = None,
        schedule: Optional[str] = None,
        start_time=None,
        end_time=None,
        query: Optional[str] = None,
        url: Optional[str] = None,
        user: Optional[str] = None,
        database: Optional[str] = None,
        db_schema: Optional[str] = None,
        warehouse: Optional[str] = None,
        **kwargs,
    ):
        attributes = attributes or {}
        if url:
            attributes["url"] = url
        if user:
            attributes["user"] = user
        if database:
            attributes["database"] = database
        if db_schema:
            attributes["db_schema"] = db_schema
        if warehouse:
            attributes["warehouse"] = warehouse
        if query:
            attributes["query"] = query

        super().__init__(
            name,
            attributes=attributes,
            key_field=key_field,
            time_field=time_field,
            schedule=schedule,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )

    def get_spark_options(self):
        spark_options = get_snowflake_spark_options(self.attributes)
        spark_options["query"] = self.attributes.get("query")
        return spark_options

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        raise mlrun.errors.MLRunRuntimeError(
            f"{type(self).__name__} supports only spark engine"
        )


class CustomSource(BaseSourceDriver):
    kind = "custom"
    support_storey = True
    support_spark = False

    def __init__(
        self,
        class_name: Optional[str] = None,
        name: str = "",
        schedule: Optional[str] = None,
        **attributes,
    ):
        attributes = attributes or {}
        attributes["class_name"] = class_name
        super().__init__(name, "", attributes, schedule=schedule)

    def to_step(self, key_field=None, time_field=None, context=None):
        attributes = copy(self.attributes)
        class_name = attributes.pop("class_name")
        class_object = get_class(class_name)
        return class_object(context=context, **attributes)


class DataFrameSource:
    """
    Reads data frame as input source for a flow.

    :parameter key_field: the column to be used as the key for events. Can be a list of keys. Defaults to None
    :parameter context: MLRun context. Defaults to None
    """

    support_storey = True

    def __init__(self, df, key_field=None, context=None, iterator=False):
        self._df = df
        if isinstance(key_field, str):
            self.key_field = [key_field]
        else:
            self.key_field = key_field
        self.context = context
        self.iterator = iterator

    def to_step(self, key_field=None, time_field=None, context=None):
        import storey

        return storey.DataframeSource(
            dfs=self._df,
            key_field=self.key_field or key_field,
            context=self.context or context,
        )

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        return self._df

    def is_iterator(self):
        return self.iterator


class OnlineSource(BaseSourceDriver):
    """online data source spec"""

    _dict_fields = [
        "kind",
        "name",
        "path",
        "attributes",
        "key_field",
        "time_field",
        "online",
        "workers",
    ]
    kind = ""

    def __init__(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        attributes: Optional[dict[str, object]] = None,
        key_field: Optional[str] = None,
        time_field: Optional[str] = None,
        workers: Optional[int] = None,
    ):
        super().__init__(name, path, attributes, key_field, time_field)
        self.online = True
        self.workers = workers

    def to_step(self, key_field=None, time_field=None, context=None):
        import storey

        source_args = self.attributes.get("source_args", {})
        explicit_ack = (
            is_explicit_ack_supported(context)
            and mlrun.mlconf.is_explicit_ack_enabled()
        )
        src_class = storey.AsyncEmitSource(
            context=context,
            key_field=self.key_field or key_field,
            full_event=True,
            explicit_ack=explicit_ack,
            **source_args,
        )

        return src_class

    def add_nuclio_trigger(self, function):
        raise mlrun.errors.MLRunInvalidArgumentError(
            "This source type is not supported with ingestion service yet"
        )

    @staticmethod
    def set_explicit_ack_mode(function: Function, **extra_arguments) -> dict[str, Any]:
        extra_arguments = extra_arguments or {}
        engine = "sync"
        if function.spec and hasattr(function.spec, "graph"):
            engine = getattr(function.spec.graph, "engine", None) or engine
        if mlrun.mlconf.is_explicit_ack_enabled() and engine == "async":
            extra_arguments["explicit_ack_mode"] = extra_arguments.get(
                "explicit_ack_mode", "explicitOnly"
            )
            extra_arguments["worker_allocation_mode"] = extra_arguments.get(
                "worker_allocation_mode", "static"
            )
        return extra_arguments


class HttpSource(OnlineSource):
    kind = "http"

    def add_nuclio_trigger(self, function):
        trigger_args = self.attributes.get("trigger_args")
        if trigger_args:
            function.with_http(**trigger_args)
        return function


class StreamSource(OnlineSource):
    kind = "v3ioStream"

    def __init__(
        self,
        name="stream",
        group="serving",
        seek_to="earliest",
        shards=1,
        retention_in_hours=24,
        extra_attributes: Optional[dict] = None,
        **kwargs,
    ):
        """
        Sets the stream source for the flow. If the stream doesn't exist it will create it.

        :param name: stream name. Default "stream"
        :param group: consumer group. Default "serving"
        :param seek_to: from where to consume the stream. Default earliest
        :param shards: number of shards in the stream. Default 1
        :param retention_in_hours: if stream doesn't exist and it will be created set retention time. Default 24h
        :param extra_attributes: additional nuclio trigger attributes (key/value dict)
        """
        attrs = {
            "group": group,
            "seek_to": seek_to,
            "shards": shards,
            "retention_in_hours": retention_in_hours,
            "extra_attributes": extra_attributes or {},
        }
        super().__init__(name, attributes=attrs, **kwargs)

    def add_nuclio_trigger(self, function):
        store, _, url = mlrun.store_manager.get_or_create_store(self.path)
        if store.kind != "v3io":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Only profiles that reference the v3io datastore can be used with StreamSource"
            )
        storage_options = store.get_storage_options()
        access_key = storage_options.get("v3io_access_key")
        endpoint, stream_path = parse_path(url)
        v3io_client = v3io.dataplane.Client(endpoint=endpoint, access_key=access_key)
        container, stream_path = split_path(stream_path)
        res = v3io_client.stream.create(
            container=container,
            stream_path=stream_path,
            shard_count=self.attributes["shards"],
            retention_period_hours=self.attributes["retention_in_hours"],
            raise_for_status=v3io.dataplane.RaiseForStatus.never,
        )
        res.raise_for_status([409, 204])
        kwargs = self.set_explicit_ack_mode(function=function)

        function.add_v3io_stream_trigger(
            url,
            self.name,
            self.attributes["group"],
            self.attributes["seek_to"],
            self.attributes["shards"],
            extra_attributes=self.attributes.get("extra_attributes", {}),
            **kwargs,
        )
        return function


class KafkaSource(OnlineSource):
    kind = "kafka"

    def __init__(
        self,
        brokers: Optional[list[str]] = None,
        topics: Optional[list[str]] = None,
        group: str = "serving",
        initial_offset: Literal["earliest", "latest"] = "earliest",
        partitions: Optional[list[int]] = None,
        sasl_user: Optional[str] = None,
        sasl_pass: Optional[str] = None,
        tls_enable: Optional[bool] = None,
        attributes: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """Sets kafka source for the flow

        :param brokers: list of broker IP addresses
        :param topics: list of topic names on which to listen.
        :param group: consumer group. Default "serving"
        :param initial_offset: from where to consume the stream. Default earliest
        :param partitions: Optional, A list of partitions numbers for which the function receives events.
        :param sasl_user: Optional, user name to use for sasl authentications
        :param sasl_pass: Optional, password to use for sasl authentications
        :param tls_enable: Optional, if set - whether to enable TLS or not.
        :param attributes: Optional, extra attributes to be passed to kafka trigger
        """
        if isinstance(topics, str):
            topics = [topics]
        if isinstance(brokers, str):
            brokers = [brokers]
        attributes = {} if attributes is None else copy(attributes)
        attributes["brokers"] = brokers
        attributes["topics"] = topics
        attributes["group"] = group
        attributes["initial_offset"] = initial_offset
        if partitions is not None:
            attributes["partitions"] = partitions

        kafka_params = mlrun.datastore.utils.KafkaParameters(attributes)

        if sasl := kafka_params.sasl(usr=sasl_user, pwd=sasl_pass):
            attributes["sasl"] = sasl

        if tls := kafka_params.tls(tls_enable=tls_enable):
            attributes["tls"] = tls

        super().__init__(attributes=attributes, **kwargs)

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        raise mlrun.MLRunInvalidArgumentError(
            "KafkaSource does not support batch processing"
        )

    def add_nuclio_trigger(self, function):
        if self.path and self.path.startswith("ds://"):
            datastore_profile = datastore_profile_read(self.path)
            extra_attributes = datastore_profile.attributes()
        else:
            extra_attributes = copy(self.attributes)
        partitions = extra_attributes.pop("partitions", None)

        extra_attributes = self.set_explicit_ack_mode(function, **extra_attributes)
        explicit_ack_mode = extra_attributes.get("explicit_ack_mode")
        extra_attributes["workerAllocationMode"] = extra_attributes.get(
            "worker_allocation_mode", "pool"
        )

        trigger_kwargs = {}

        if "max_workers" in extra_attributes:
            trigger_kwargs = {"max_workers": extra_attributes.pop("max_workers")}

        trigger = KafkaTrigger(
            brokers=extra_attributes.pop("brokers"),
            topics=extra_attributes.pop("topics"),
            partitions=partitions,
            consumer_group=extra_attributes.pop("group"),
            initial_offset=extra_attributes.pop("initial_offset"),
            explicit_ack_mode=explicit_ack_mode,
            extra_attributes=extra_attributes,
            **trigger_kwargs,
        )
        function = function.add_trigger("kafka", trigger)

        # ML-5499
        bug_fix_version = "1.12.10"
        if config.nuclio_version and semver.VersionInfo.parse(
            config.nuclio_version
        ) < semver.VersionInfo.parse(bug_fix_version):
            warnings.warn(
                f"Detected nuclio version {config.nuclio_version}, which is older "
                f"than {bug_fix_version}. Forcing number of replicas of 1 in function '{function.metadata.name}'. "
                f"To resolve this, please upgrade Nuclio."
            )
            function.spec.min_replicas = 1
            function.spec.max_replicas = 1

        return function

    def to_spark_df(self, session, named_view=False, time_field=None, columns=None):
        raise NotImplementedError(
            "Conversion of a source of type 'KafkaSource' "
            "to a Spark dataframe is not possible, as this operation is not supported by Spark"
        )

    def create_topics(
        self,
        num_partitions: int = 4,
        replication_factor: int = 1,
        topics: Optional[list[str]] = None,
    ):
        """
        Create Kafka topics with the specified number of partitions and replication factor.

        :param num_partitions:      number of partitions for the topics
        :param replication_factor:  replication factor for the topics
        :param topics:              list of topic names to create, if None,
                                    the topics will be taken from the source attributes
        """
        from kafka.admin import KafkaAdminClient, NewTopic

        brokers = self.attributes.get("brokers")
        if not brokers:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "brokers must be specified in the KafkaSource attributes"
            )
        topics = topics or self.attributes.get("topics")
        if not topics:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "topics must be specified in the KafkaSource attributes"
            )
        new_topics = [
            NewTopic(topic, num_partitions, replication_factor) for topic in topics
        ]

        kafka_admin_kwargs = {}
        kafka_admin_kwargs = mlrun.datastore.utils.KafkaParameters(
            self.attributes
        ).admin()

        kafka_admin = KafkaAdminClient(bootstrap_servers=brokers, **kafka_admin_kwargs)
        try:
            kafka_admin.create_topics(new_topics)
        finally:
            kafka_admin.close()
        logger.info(
            "Kafka topics created successfully",
            topics=topics,
            num_partitions=num_partitions,
            replication_factor=replication_factor,
        )


class SQLSource(BaseSourceDriver):
    kind = "sqldb"
    support_storey = True
    support_spark = False

    def __init__(
        self,
        name: str = "",
        chunksize: Optional[int] = None,
        key_field: Optional[str] = None,
        time_field: Optional[str] = None,
        schedule: Optional[str] = None,
        start_time: Optional[Union[datetime, str]] = None,
        end_time: Optional[Union[datetime, str]] = None,
        db_url: Optional[str] = None,
        table_name: Optional[str] = None,
        spark_options: Optional[dict] = None,
        parse_dates: Optional[list[str]] = None,
        **kwargs,
    ):
        """
        Reads SqlDB as input source for a flow.
        example::
            db_url = "mysql+pymysql://<username>:<password>@<host>:<port>/<db_name>"
            source = SQLSource(
                table_name='source_name', db_url=db_url, key_field='key'
            )
        :param name:            source name
        :param chunksize:       number of rows per chunk (default large single chunk)
        :param key_field:       the column to be used as the key for the collection.
        :param time_field:      the column to be parsed as timestamp for events. Defaults to None
        :param start_time:      filters out data before this time
        :param end_time:        filters out data after this time
        :param schedule:        string to configure scheduling of the ingestion job.
                                For example '*/30 * * * *' will
                                cause the job to run every 30 minutes
        :param db_url:         url string connection to sql database.
                                If not set, the MLRUN_SQL__URL environment variable will be used.
        :param table_name:      the name of the collection to access,
                                from the current database
        :param spark_options:   additional spark read options
        :param parse_dates :    all the field to be parsed as timestamp.
        """
        db_url = db_url or mlrun.mlconf.sql.url
        if db_url is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "cannot specify without db_path arg or secret MLRUN_SQL__URL"
            )
        if time_field:
            if parse_dates:
                parse_dates.append(time_field)
            else:
                parse_dates = [time_field]
        attrs = {
            "chunksize": chunksize,
            "spark_options": spark_options,
            "table_name": table_name,
            "db_path": db_url,
            "parse_dates": parse_dates,
        }
        attrs = {key: value for key, value in attrs.items() if value is not None}
        super().__init__(
            name,
            attributes=attrs,
            key_field=key_field,
            time_field=time_field,
            schedule=schedule,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )

    def to_dataframe(
        self,
        columns=None,
        df_module=None,
        entities=None,
        start_time=None,
        end_time=None,
        time_field=None,
        additional_filters=None,
    ):
        import sqlalchemy as sqlalchemy

        mlrun.utils.helpers.additional_filters_warning(
            additional_filters, self.__class__
        )
        db_path = self.attributes.get("db_path")
        table_name = self.attributes.get("table_name")
        parse_dates = self.attributes.get("parse_dates")
        time_field = time_field or self.time_field
        start_time = start_time or self.start_time
        end_time = end_time or self.end_time
        if table_name and db_path:
            engine = sqlalchemy.create_engine(db_path)
            query, parse_dates = _generate_sql_query_with_time_filter(
                table_name=table_name,
                engine=engine,
                time_column=time_field,
                parse_dates=parse_dates,
                start_time=start_time,
                end_time=end_time,
            )
            with engine.connect() as con:
                return pd.read_sql(
                    query,
                    con=con,
                    chunksize=self.attributes.get("chunksize"),
                    parse_dates=parse_dates,
                    columns=columns,
                )
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "table_name and db_name args must be specified"
            )

    def to_step(self, key_field=None, time_field=None, context=None):
        import storey

        attributes = self.attributes or {}
        if context:
            attributes["context"] = context

        return storey.SQLSource(
            key_field=self.key_field or key_field,
            time_field=self.time_field or time_field,
            end_filter=self.end_time,
            start_filter=self.start_time,
            filter_column=self.time_field or time_field,
            **attributes,
        )
        pass

    def is_iterator(self):
        return bool(self.attributes.get("chunksize"))


# map of sources (exclude DF source which is not serializable)
source_kind_to_driver = {
    "": BaseSourceDriver,
    CSVSource.kind: CSVSource,
    ParquetSource.kind: ParquetSource,
    HttpSource.kind: HttpSource,
    StreamSource.kind: StreamSource,
    KafkaSource.kind: KafkaSource,
    CustomSource.kind: CustomSource,
    BigQuerySource.kind: BigQuerySource,
    SnowflakeSource.kind: SnowflakeSource,
    SQLSource.kind: SQLSource,
}



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1


class PipelinesPagination(str):
    default_page_size = 200
    # https://github.com/kubeflow/pipelines/blob/master/backend/src/apiserver/list/list.go#L363
    max_page_size = 200


class PipelinesOutput(pydantic.v1.BaseModel):
    # use the format query param to control what is returned
    runs: list[typing.Union[dict, str]]
    total_size: int
    next_page_token: typing.Optional[str]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional

from torch import Tensor
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader

from ..utils import PyTorchTypes


class Callback(ABC):
    """
    Abstract class for a callback to MLRun's pytorch framework package. Each callback must implement this class so it
    could be used in MLRun's PyTorch interface. If you wish to implement custom training / evaluation, you should
    consider using the 'CallbacksHandler' class.
    """

    class _ObjectKeys:
        """
        Keys for the objects dictionary of the callback. Each callback can choose what to store in its objects
        dictionary.
        """

        MODEL = "model"
        TRAINING_SET = "training_set"
        VALIDATION_SET = "validation_set"
        LOSS_FUNCTION = "loss_function"
        OPTIMIZER = "optimizer"
        METRIC_FUNCTIONS = "metric_functions"
        SCHEDULER = "scheduler"

    @abstractmethod
    def __init__(self):
        """
        Initialize the callback with an empty objects dictionary. The objects should be registered on setup.
        """
        self._objects = {}

    def on_horovod_check(self, rank: int) -> bool:
        """
        Check whether this callback is fitting to run by the given horovod rank (worker).

        :param rank: The horovod rank (worker) id.

        :return: True if the callback is ok to run on this rank and false if not.
        """
        pass

    def on_setup(
        self,
        model: Module = None,
        training_set: DataLoader = None,
        validation_set: DataLoader = None,
        loss_function: Module = None,
        optimizer: Optimizer = None,
        metric_functions: Optional[list[PyTorchTypes.MetricFunctionType]] = None,
        scheduler=None,
    ):
        """
        Basic setup command, storing all the given objects in the callback's objects dictionary.

        :param model:            The model to be stored in this callback.
        :param training_set:     The training set to be stored in this callback.
        :param validation_set:   The validation set to be stored in this callback.
        :param loss_function:    The loss function to be stored in this callback.
        :param optimizer:        The optimizer to be stored in this callback.
        :param metric_functions: The metric functions to be stored in this callback.
        :param scheduler:        The scheduler to be stored in this callback.
        """
        self._objects[self._ObjectKeys.MODEL] = model
        self._objects[self._ObjectKeys.TRAINING_SET] = training_set
        self._objects[self._ObjectKeys.VALIDATION_SET] = validation_set
        self._objects[self._ObjectKeys.LOSS_FUNCTION] = loss_function
        self._objects[self._ObjectKeys.OPTIMIZER] = optimizer
        self._objects[self._ObjectKeys.METRIC_FUNCTIONS] = metric_functions
        self._objects[self._ObjectKeys.SCHEDULER] = scheduler

    def on_run_begin(self):
        """
        After the run begins, this method will be called.
        """
        pass

    def on_run_end(self):
        """
        Before the run ends, this method will be called.
        """
        pass

    def on_epoch_begin(self, epoch: int):
        """
        After the epoch begins, this method will be called.

        :param epoch: The epoch that is about to begin.
        """
        pass

    def on_epoch_end(self, epoch: int) -> bool:
        """
        Before the epoch ends, this method will be called.

        :param epoch: The epoch that has just ended.

        :return Can optionally return a boolean value indicating whether or not to continue the training process.
        """
        return True

    def on_train_begin(self):
        """
        After the training of the current epoch begins, this method will be called.
        """
        pass

    def on_train_end(self) -> bool:
        """
        Before the training of the current epoch ends, this method will be called.

        :return Can optionally return a boolean value indicating whether or not to continue the training process.
        """
        return True

    def on_validation_begin(self):
        """
        After the validation (in a training case it will be per epoch) begins, this method will be called.
        """
        pass

    def on_validation_end(
        self, loss_value: PyTorchTypes.MetricValueType, metric_values: list[float]
    ) -> bool:
        """
        Before the validation (in a training case it will be per epoch) ends, this method will be called.

        :param loss_value:    The loss summary of this validation.
        :param metric_values: The metrics summaries of this validation.

        :return Can optionally return a boolean value indicating whether or not to continue the training / evaluation
                process.
        """
        return True

    def on_train_batch_begin(self, batch: int, x: Tensor, y_true: Tensor):
        """
        After the training of the given batch begins, this method will be called.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input of the current batch.
        :param y_true: The ground truth value of the current batch.
        """
        pass

    def on_train_batch_end(
        self, batch: int, x: Tensor, y_pred: Tensor, y_true: Tensor
    ) -> bool:
        """
        Before the training of the given batch ends, this method will be called.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input of the current batch.
        :param y_pred: The prediction (output) of the model for this batch's input ('x').
        :param y_true: The ground truth value of the current batch.

        :return Can optionally return a boolean value indicating whether or not to continue the training process.
        """
        return True

    def on_validation_batch_begin(self, batch: int, x: Tensor, y_true: Tensor):
        """
        After the validation of the given batch begins, this method will be called.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input of the current batch.
        :param y_true: The ground truth value of the current batch.
        """
        pass

    def on_validation_batch_end(
        self, batch: int, x: Tensor, y_pred: Tensor, y_true: Tensor
    ) -> bool:
        """
        Before the validation of the given batch ends, this method will be called.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input of the current batch.
        :param y_pred: The prediction (output) of the model for this batch's input ('x').
        :param y_true: The ground truth value of the current batch.

        :return Can optionally return a boolean value indicating whether or not to continue the training / evaluation
                process.
        """
        return True

    def on_inference_begin(self, x: Tensor):
        """
        Before the inference of the current batch sample into the model, this method will be called to process the
        input.

        :param x: The input of the current batch.
        """
        pass

    def on_inference_end(self, y_pred: Tensor, y_true: Tensor):
        """
        After the inference of the current batch sample, this method will be called to process the output along side the
        current batch ground truth.

        :param y_pred: The prediction (output) of the model for this batch's input ('x').
        :param y_true: The ground truth value of the current batch.
        """
        pass

    def on_train_loss_begin(self):
        """
        Before the training calculation of the loss, this method will be called.
        """
        pass

    def on_train_loss_end(self, loss_value: PyTorchTypes.MetricValueType):
        """
        After the training calculation of the loss, this method will be called.

        :param loss_value: The recent loss value calculated during training.
        """
        pass

    def on_validation_loss_begin(self):
        """
        Before the validating calculation of the loss, this method will be called.
        """
        pass

    def on_validation_loss_end(self, loss_value: PyTorchTypes.MetricValueType):
        """
        After the validating calculation of the loss, this method will be called.

        :param loss_value: The recent loss value calculated during validation.
        """
        pass

    def on_train_metrics_begin(self):
        """
        Before the training calculation of the metrics, this method will be called.
        """
        pass

    def on_train_metrics_end(self, metric_values: list[PyTorchTypes.MetricValueType]):
        """
        After the training calculation of the metrics, this method will be called.

        :param metric_values: The recent metric values calculated during training.
        """
        pass

    def on_validation_metrics_begin(self):
        """
        Before the validating calculation of the metrics, this method will be called.
        """
        pass

    def on_validation_metrics_end(
        self, metric_values: list[PyTorchTypes.MetricValueType]
    ):
        """
        After the validating calculation of the metrics, this method will be called.

        :param metric_values: The recent metric values calculated during validation.
        """
        pass

    def on_backward_begin(self):
        """
        Before the backward propagation of the loss function, this method will be called.
        """
        pass

    def on_backward_end(self):
        """
        After the backward propagation of the loss function, this method will be called.
        """
        pass

    def on_optimizer_step_begin(self):
        """
        Before the optimizer 'step' and 'zero_grad' methods are called, this method will be called.
        """
        pass

    def on_optimizer_step_end(self):
        """
        After the optimizer 'step' and 'zero_grad' methods are called, this method will be called.
        """
        pass

    def on_scheduler_step_begin(self):
        """
        Before the scheduler 'step' method is called, this method will be called.
        """
        pass

    def on_scheduler_step_end(self):
        """
        After the scheduler 'step' method is called, this method will be called.
        """
        pass

    def on_call_check(self) -> bool:
        """
        Before the callbacks handler is calling its callbacks, this method will be called to know if this callback
        should run. For example, in case of multiprocessing, logging should happen only for callbacks who are called
        from worker 0. The worker id check should be done here.

        :return: True if the call is ok to run and false if not.
        """
        return True



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import lightgbm as lgb
import numpy as np
import pandas as pd

import mlrun.errors

from .._ml_common import AlgorithmFunctionality, MLTypes, MLUtils


class LGBMTypes(MLTypes):
    """
    Typing hints for the LightGBM framework.
    """

    # A union of all LightGBM model base classes:
    ModelType = Union[lgb.LGBMModel, lgb.Booster]

    # A type for all the supported dataset types:
    DatasetType = Union[MLTypes.DatasetType, lgb.Dataset]

    # An evaluation result as packaged by the training in LightGBM:
    EvaluationResultType = Union[
        tuple[str, str, float, bool],  # As packaged in `lightgbm.train`
        tuple[str, str, float, bool, float],  # As packaged in `lightgbm.cv`
    ]

    # Detailed type for the named tuple `CallbackEnv` passed during LightGBM's training for the callbacks:
    CallbackEnvType = tuple[
        lgb.Booster, dict, int, int, int, list[EvaluationResultType]
    ]


class LGBMUtils(MLUtils):
    """
    Utilities functions for the LightGBM framework.
    """

    @staticmethod
    def to_array(dataset: LGBMTypes.DatasetType) -> np.ndarray:
        """
        Convert the given dataset to np.ndarray.

        :param dataset: The dataset to convert. Must be one of {lgb.Dataset, pd.DataFrame, pd.Series,
                        scipy.sparse.base.spmatrix, list, tuple, dict}.

        :return: The dataset as a ndarray.

        :raise MLRunInvalidArgumentError: If the dataset type is not supported.
        """
        if isinstance(dataset, lgb.Dataset):
            x = LGBMUtils.to_array(dataset=dataset.data)
            if dataset.label is None:
                return x
            y = LGBMUtils.to_array(dataset=dataset.label)
            return LGBMUtils.to_array(LGBMUtils.concatenate_x_y(x=x, y=y)[0])
        try:
            return MLUtils.to_array(dataset=dataset)
        except mlrun.errors.MLRunInvalidArgumentError:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Could not convert the given dataset into a numpy ndarray. Supporting conversion from: "
                f"{LGBMUtils.get_union_typehint_string(LGBMTypes.DatasetType)}. The given dataset was of type: "
                f"'{type(dataset)}'"
            )

    @staticmethod
    def to_dataframe(dataset: LGBMTypes.DatasetType) -> pd.DataFrame:
        """
        Convert the given dataset to pd.DataFrame.

        :param dataset: The dataset to convert. Must be one of {lgb.Dataset, np.ndarray, pd.Series,
                        scipy.sparse.base.spmatrix, list, tuple, dict}.

        :return: The dataset as a DataFrame.

        :raise MLRunInvalidArgumentError: If the dataset type is not supported.
        """
        if isinstance(dataset, lgb.Dataset):
            x = LGBMUtils.to_dataframe(dataset=dataset.data)
            if dataset.label is None:
                return x
            y = LGBMUtils.to_dataframe(dataset=dataset.label)
            return LGBMUtils.concatenate_x_y(x=x, y=y)[0]
        try:
            return MLUtils.to_dataframe(dataset=dataset)
        except mlrun.errors.MLRunInvalidArgumentError:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Could not convert the given dataset into a pandas DataFrame. Supporting conversion from: "
                f"{LGBMUtils.get_union_typehint_string(LGBMTypes.DatasetType)}. The given dataset was of type: "
                f"'{type(dataset)}'"
            )

    @staticmethod
    def get_algorithm_functionality(
        model: MLTypes.ModelType = None,
        y: MLTypes.DatasetType = None,
        objective: Optional[str] = None,
    ) -> AlgorithmFunctionality:
        """
        Get the algorithm functionality of the LightGBM model. If SciKit-Learn API is used, pass the LGBBMModel and a y
        sample. Otherwise, training API is used, so pass the objective of the params dictionary.

        The objectives here are taken from the official docs of LightGBBM at:
        https://lightgbm.readthedocs.io/en/latest/Parameters.html#core-parameters

        :param model:     The model to check if its a regression model or a classification model (SciKit-Learn API).
        :param y:         The ground truth values to check if its multiclass and / or multi output (SciKit-Learn API).
        :param objective: The objective string (Training API).

        :return: The objective's algorithm functionality.
        """
        # Check if LightGBM is being used with SciKit-Learn API:
        if objective is None:
            return super().get_algorithm_functionality(model=model, y=y)

        # Declare the conversion map according to the LightGBM docs:
        objective_to_algorithm_functionality_map = {
            # regression application:
            "regression": AlgorithmFunctionality.REGRESSION,
            "regression_l2": AlgorithmFunctionality.REGRESSION,
            "l2": AlgorithmFunctionality.REGRESSION,
            "mean_squared_error": AlgorithmFunctionality.REGRESSION,
            "mse": AlgorithmFunctionality.REGRESSION,
            "l2_root": AlgorithmFunctionality.REGRESSION,
            "root_mean_squared_error": AlgorithmFunctionality.REGRESSION,
            "rmse": AlgorithmFunctionality.REGRESSION,
            "regression_l1": AlgorithmFunctionality.REGRESSION,
            "l1": AlgorithmFunctionality.REGRESSION,
            "mean_absolute_error": AlgorithmFunctionality.REGRESSION,
            "mae": AlgorithmFunctionality.REGRESSION,
            "huber": AlgorithmFunctionality.REGRESSION,
            "fair": AlgorithmFunctionality.REGRESSION,
            "poisson": AlgorithmFunctionality.REGRESSION,
            "quantile": AlgorithmFunctionality.REGRESSION,
            "mape": AlgorithmFunctionality.REGRESSION,
            "mean_absolute_percentage_error": AlgorithmFunctionality.REGRESSION,
            "gamma": AlgorithmFunctionality.REGRESSION,
            "tweedie": AlgorithmFunctionality.REGRESSION,
            # binary classification application:
            "binary": AlgorithmFunctionality.BINARY_CLASSIFICATION,
            # multi-class classification application:
            "multiclass": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "softmax": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "multiclassova": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "multiclass_ova": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "ova": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "ovr": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            # cross-entropy application
            "cross_entropy": AlgorithmFunctionality.BINARY_CLASSIFICATION,
            "xentropy": AlgorithmFunctionality.BINARY_CLASSIFICATION,
            "cross_entropy_lambda": AlgorithmFunctionality.BINARY_CLASSIFICATION,
            "xentlambda": AlgorithmFunctionality.BINARY_CLASSIFICATION,
            # ranking application
            "lambdarank": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "rank_xendcg": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "xendcg": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "xe_ndcg": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "xe_ndcg_mart": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
            "xendcg_mart": AlgorithmFunctionality.MULTICLASS_CLASSIFICATION,
        }

        # Return unknown if the objective is not in the map and otherwise return its functionality:
        if objective not in objective_to_algorithm_functionality_map:
            raise AlgorithmFunctionality.UNKNOWN
        return objective_to_algorithm_functionality_map[objective]



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from dataclasses import dataclass
from io import StringIO
from typing import Optional, Union

import taosws

import mlrun.common.schemas.model_monitoring as mm_schemas
import mlrun.common.types

_MODEL_MONITORING_DATABASE = "mlrun_model_monitoring"


class _TDEngineColumnType:
    def __init__(self, data_type: str, length: Optional[int] = None):
        self.data_type = data_type
        self.length = length

    def values_to_column(self, values):
        raise NotImplementedError()

    def __str__(self):
        if self.length is not None:
            return f"{self.data_type}({self.length})"
        else:
            return self.data_type


class _TDEngineColumn(mlrun.common.types.StrEnum):
    TIMESTAMP = _TDEngineColumnType("TIMESTAMP")
    FLOAT = _TDEngineColumnType("FLOAT")
    INT = _TDEngineColumnType("INT")
    BINARY_40 = _TDEngineColumnType("BINARY", 40)
    BINARY_64 = _TDEngineColumnType("BINARY", 64)
    BINARY_1000 = _TDEngineColumnType("BINARY", 1000)


def values_to_column(values, column_type):
    if column_type == _TDEngineColumn.TIMESTAMP:
        timestamps = [round(timestamp.timestamp() * 1000) for timestamp in values]
        return taosws.millis_timestamps_to_column(timestamps)
    if column_type == _TDEngineColumn.FLOAT:
        return taosws.floats_to_column(values)
    if column_type == _TDEngineColumn.INT:
        return taosws.ints_to_column(values)
    if column_type == _TDEngineColumn.BINARY_40:
        return taosws.binary_to_column(values)
    if column_type == _TDEngineColumn.BINARY_64:
        return taosws.binary_to_column(values)
    if column_type == _TDEngineColumn.BINARY_1000:
        return taosws.binary_to_column(values)

    raise mlrun.errors.MLRunInvalidArgumentError(
        f"unsupported column type '{column_type}'"
    )


@dataclass
class TDEngineSchema:
    """
    A class to represent a supertable schema in TDengine. Using this schema, you can generate the relevant queries to
    create, insert, delete and query data from TDengine. At the moment, there are 3 schemas: AppResultTable,
    Metrics, and Predictions.
    """

    def __init__(
        self,
        super_table: str,
        columns: dict[str, _TDEngineColumn],
        tags: dict[str, str],
        project: str,
        database: Optional[str] = None,
    ):
        self.super_table = f"{super_table}_{project.replace('-', '_')}"
        self.columns = columns
        self.tags = tags
        self.database = database or _MODEL_MONITORING_DATABASE

    def _create_super_table_query(self) -> str:
        columns = ", ".join(f"{col} {val}" for col, val in self.columns.items())
        tags = ", ".join(f"{col} {val}" for col, val in self.tags.items())
        return f"CREATE STABLE if NOT EXISTS {self.database}.{self.super_table} ({columns}) TAGS ({tags});"

    def _create_subtable_sql(
        self,
        subtable: str,
        values: dict[str, Union[str, int, float, datetime.datetime]],
    ) -> str:
        try:
            tags = ", ".join(f"'{values[val]}'" for val in self.tags)
        except KeyError:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"values must contain all tags: {self.tags.keys()}"
            )
        return f"CREATE TABLE if NOT EXISTS {self.database}.{subtable} USING {self.super_table} TAGS ({tags});"

    def _delete_subtable_query(
        self,
        subtable: str,
        values: dict[str, Union[str, int, float, datetime.datetime]],
    ) -> str:
        values = " AND ".join(
            f"{val} LIKE '{values[val]}'" for val in self.tags if val in values
        )
        if not values:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"values must contain at least one tag: {self.tags.keys()}"
            )
        return f"DELETE FROM {self.database}.{subtable} WHERE {values};"

    def drop_subtable_query(self, subtable: str) -> str:
        return f"DROP TABLE if EXISTS {self.database}.`{subtable}`;"

    def drop_supertable_query(self) -> str:
        return f"DROP STABLE if EXISTS {self.database}.{self.super_table};"

    def _get_subtables_query_by_tag(
        self,
        filter_tag: str,
        filter_values: list[str],
        operator: str = "OR",
    ) -> str:
        if filter_tag not in self.tags:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"`filter_tag` must be one of the tags: {self.tags.keys()}"
            )

        values = f" {operator} ".join(
            f"{filter_tag} LIKE '{val}'" for val in filter_values
        )
        return self._get_tables_query_by_condition(values)

    def _get_tables_query_by_condition(self, condition: str) -> str:
        return f"SELECT DISTINCT TBNAME FROM {self.database}.{self.super_table} WHERE {condition};"

    @staticmethod
    def _get_records_query(
        table: str,
        start: datetime.datetime,
        end: datetime.datetime,
        columns_to_filter: Optional[list[str]] = None,
        filter_query: Optional[str] = None,
        interval: Optional[str] = None,
        limit: int = 0,
        agg_funcs: Optional[list] = None,
        sliding_window_step: Optional[str] = None,
        timestamp_column: str = "time",
        database: str = _MODEL_MONITORING_DATABASE,
        group_by: Optional[Union[list[str], str]] = None,
        preform_agg_funcs_columns: Optional[list[str]] = None,
        order_by: Optional[str] = None,
        desc: Optional[bool] = None,
        partition_by: Optional[str] = None,
    ) -> str:
        if agg_funcs and not columns_to_filter:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`columns_to_filter` must be provided when using aggregate functions"
            )

        # if aggregate function or interval is provided, the other must be provided as well
        if interval and not agg_funcs:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`agg_funcs` must be provided when using interval"
            )
        if partition_by and not agg_funcs:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`agg_funcs` must be provided when using partition by"
            )
        if sliding_window_step and not interval:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`interval` must be provided when using sliding window"
            )
        if group_by and not agg_funcs:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "aggregate functions must be provided when using group by"
            )
        if desc and not order_by:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "`order_by` must be provided when using descending"
            )

        with StringIO() as query:
            query.write("SELECT ")
            if interval:
                query.write("_wstart, _wend, ")
            if agg_funcs:
                preform_agg_funcs_columns = (
                    columns_to_filter
                    if preform_agg_funcs_columns is None
                    else preform_agg_funcs_columns
                )
                query.write(
                    ", ".join(
                        [
                            f"{a}({col})"
                            if col.upper()
                            in map(
                                str.upper, preform_agg_funcs_columns
                            )  # Case-insensitive check
                            else f"{col}"
                            for a in agg_funcs
                            for col in columns_to_filter
                        ]
                    )
                )
            elif columns_to_filter:
                query.write(", ".join(columns_to_filter))
            else:
                query.write("*")
            query.write(f" FROM {database}.{table}")

            if any([filter_query, start, end]):
                query.write(" WHERE ")
                if filter_query:
                    query.write(f"{filter_query} AND ")
                if start:
                    query.write(f"{timestamp_column} >= '{start}' AND ")
                if end:
                    query.write(f"{timestamp_column} <= '{end}'")
            if group_by:
                if isinstance(group_by, list):
                    group_by = ", ".join(group_by)
                query.write(f" GROUP BY {group_by}")
            if partition_by:
                query.write(f" PARTITION BY {partition_by}")
            if order_by:
                desc = " DESC" if desc else ""
                query.write(f" ORDER BY {order_by}{desc}")
            if interval:
                query.write(f" INTERVAL({interval})")
            if sliding_window_step:
                query.write(f" SLIDING({sliding_window_step})")
            if limit:
                query.write(f" LIMIT {limit}")
            query.write(";")
            return query.getvalue()


@dataclass
class AppResultTable(TDEngineSchema):
    def __init__(self, project: str, database: Optional[str] = None):
        super_table = mm_schemas.TDEngineSuperTables.APP_RESULTS
        columns = {
            mm_schemas.WriterEvent.END_INFER_TIME: _TDEngineColumn.TIMESTAMP,
            mm_schemas.WriterEvent.START_INFER_TIME: _TDEngineColumn.TIMESTAMP,
            mm_schemas.ResultData.RESULT_VALUE: _TDEngineColumn.FLOAT,
            mm_schemas.ResultData.RESULT_STATUS: _TDEngineColumn.INT,
            mm_schemas.ResultData.RESULT_EXTRA_DATA: _TDEngineColumn.BINARY_1000,
        }
        tags = {
            mm_schemas.WriterEvent.ENDPOINT_ID: _TDEngineColumn.BINARY_64,
            mm_schemas.WriterEvent.APPLICATION_NAME: _TDEngineColumn.BINARY_64,
            mm_schemas.ResultData.RESULT_NAME: _TDEngineColumn.BINARY_64,
            mm_schemas.ResultData.RESULT_KIND: _TDEngineColumn.INT,
        }
        super().__init__(
            super_table=super_table,
            columns=columns,
            tags=tags,
            database=database,
            project=project,
        )


@dataclass
class Metrics(TDEngineSchema):
    def __init__(self, project: str, database: Optional[str] = None):
        super_table = mm_schemas.TDEngineSuperTables.METRICS
        columns = {
            mm_schemas.WriterEvent.END_INFER_TIME: _TDEngineColumn.TIMESTAMP,
            mm_schemas.WriterEvent.START_INFER_TIME: _TDEngineColumn.TIMESTAMP,
            mm_schemas.MetricData.METRIC_VALUE: _TDEngineColumn.FLOAT,
        }
        tags = {
            mm_schemas.WriterEvent.ENDPOINT_ID: _TDEngineColumn.BINARY_64,
            mm_schemas.WriterEvent.APPLICATION_NAME: _TDEngineColumn.BINARY_64,
            mm_schemas.MetricData.METRIC_NAME: _TDEngineColumn.BINARY_64,
        }
        super().__init__(
            super_table=super_table,
            columns=columns,
            tags=tags,
            database=database,
            project=project,
        )


@dataclass
class Predictions(TDEngineSchema):
    def __init__(self, project: str, database: Optional[str] = None):
        super_table = mm_schemas.TDEngineSuperTables.PREDICTIONS
        columns = {
            mm_schemas.EventFieldType.TIME: _TDEngineColumn.TIMESTAMP,
            mm_schemas.EventFieldType.LATENCY: _TDEngineColumn.FLOAT,
            mm_schemas.EventKeyMetrics.CUSTOM_METRICS: _TDEngineColumn.BINARY_1000,
            mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT: _TDEngineColumn.FLOAT,
            mm_schemas.EventFieldType.EFFECTIVE_SAMPLE_COUNT: _TDEngineColumn.INT,
        }
        tags = {
            mm_schemas.WriterEvent.ENDPOINT_ID: _TDEngineColumn.BINARY_64,
        }
        super().__init__(
            super_table=super_table,
            columns=columns,
            tags=tags,
            database=database,
            project=project,
        )


@dataclass
class Errors(TDEngineSchema):
    def __init__(self, project: str, database: Optional[str] = None):
        super_table = mm_schemas.TDEngineSuperTables.ERRORS
        columns = {
            mm_schemas.EventFieldType.TIME: _TDEngineColumn.TIMESTAMP,
            mm_schemas.EventFieldType.MODEL_ERROR: _TDEngineColumn.BINARY_1000,
        }
        tags = {
            mm_schemas.WriterEvent.ENDPOINT_ID: _TDEngineColumn.BINARY_64,
            mm_schemas.EventFieldType.ERROR_TYPE: _TDEngineColumn.BINARY_64,
        }
        super().__init__(
            super_table=super_table,
            columns=columns,
            tags=tags,
            database=database,
            project=project,
        )



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import traceback
from collections import OrderedDict
from datetime import datetime
from typing import Any, Optional, Union

import mlrun.common.schemas
import mlrun.common.schemas.alert as alert_objects
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.model_monitoring.helpers
import mlrun.platforms.iguazio
from mlrun.serving import GraphContext
from mlrun.serving.utils import StepToDict
from mlrun.utils import logger

from .base import _serialize_context_and_result
from .context import MonitoringApplicationContext
from .results import (
    ModelMonitoringApplicationMetric,
    ModelMonitoringApplicationResult,
    _ModelMonitoringApplicationStats,
)


class _PushToMonitoringWriter(StepToDict):
    kind = "monitoring_application_stream_pusher"

    def __init__(self, project: str) -> None:
        """
        Class for pushing application results to the monitoring writer stream.

        :param project: Project name.
        """
        self.project = project
        self._output_stream = None

    def do(
        self,
        event: tuple[
            list[
                Union[
                    ModelMonitoringApplicationResult,
                    ModelMonitoringApplicationMetric,
                    _ModelMonitoringApplicationStats,
                ]
            ],
            MonitoringApplicationContext,
        ],
    ) -> None:
        """
        Push application results to the monitoring writer stream.

        :param event: Monitoring result(s) to push and the original event from the controller.
        """
        application_results, application_context = event

        writer_events = [
            _serialize_context_and_result(context=application_context, result=result)
            for result in application_results
        ]

        logger.debug("Pushing data to output stream", writer_events=str(writer_events))
        self.output_stream.push(
            writer_events, partition_key=application_context.endpoint_id
        )
        logger.debug("Pushed data to output stream successfully")

    @property
    def output_stream(
        self,
    ) -> Union[
        mlrun.platforms.iguazio.OutputStream, mlrun.platforms.iguazio.KafkaOutputStream
    ]:
        if self._output_stream is None:
            self._output_stream = mlrun.model_monitoring.helpers.get_output_stream(
                project=self.project,
                function_name=mm_constants.MonitoringFunctionNames.WRITER,
            )
        return self._output_stream


class _PrepareMonitoringEvent(StepToDict):
    MAX_MODEL_ENDPOINTS: int = 1500

    def __init__(self, context: GraphContext, application_name: str) -> None:
        """
        Class for preparing the application event for the application step.

        :param application_name: Application name.
        """
        self.graph_context = context
        _ = self.graph_context.project_obj  # Ensure project exists
        self.application_name = application_name
        self.model_endpoints: OrderedDict[str, mlrun.common.schemas.ModelEndpoint] = (
            collections.OrderedDict()
        )
        self.feature_sets: dict[str, mlrun.common.schemas.FeatureSet] = {}

    def do(self, event: dict[str, Any]) -> MonitoringApplicationContext:
        """
        Prepare the application event for the application step.

        :param event: Application event.
        :return: Application context.
        """
        endpoint_id = event.get(mm_constants.ApplicationEvent.ENDPOINT_ID)
        endpoint_updated = datetime.fromisoformat(
            event.get(mm_constants.ApplicationEvent.ENDPOINT_UPDATED)
        )
        if (
            endpoint_id in self.model_endpoints
            and endpoint_updated != self.model_endpoints[endpoint_id].metadata.updated
        ):
            logger.debug(
                "Updated endpoint removing endpoint from cash",
                new_updated=endpoint_updated.isoformat(),
                old_updated=self.model_endpoints[
                    endpoint_id
                ].metadata.updated.isoformat(),
            )
            self.model_endpoints.pop(endpoint_id)

        application_context = MonitoringApplicationContext._from_graph_ctx(
            application_name=self.application_name,
            event=event,
            model_endpoint_dict=self.model_endpoints,
            graph_context=self.graph_context,
            feature_sets_dict=self.feature_sets,
        )

        self.model_endpoints.setdefault(
            application_context.endpoint_id, application_context.model_endpoint
        )
        self.feature_sets.setdefault(
            application_context.endpoint_id, application_context.feature_set
        )
        # every used endpoint goes to first location allowing to pop last used:
        self.model_endpoints.move_to_end(application_context.endpoint_id, last=False)
        if len(self.model_endpoints) > self.MAX_MODEL_ENDPOINTS:
            removed_endpoint_id, _ = self.model_endpoints.popitem(
                last=True
            )  # Removing the LRU endpoint
            self.feature_sets.pop(removed_endpoint_id, None)
            logger.debug(
                "Exceeded maximum number of model endpoints removing the LRU from cash",
                endpoint_id=removed_endpoint_id,
            )

        return application_context


class _ApplicationErrorHandler(StepToDict):
    def __init__(self, project: str, name: Optional[str] = None):
        self.project = project
        self.name = name or "ApplicationErrorHandler"

    def do(self, event):
        """
        Handle model monitoring application error. This step will generate an event, describing the error.

        :param event: Application event.
        """

        error_data = {
            "Endpoint ID": event.body.endpoint_id,
            "Application Class": event.body.application_name,
            "Error": "".join(
                traceback.format_exception(
                    None, value=event.error, tb=event.error.__traceback__
                )
            ),
            "Timestamp": event.timestamp,
        }
        logger.error("Error in application step", **error_data)

        error_data["Error"] = event.error

        event_data = alert_objects.Event(
            kind=alert_objects.EventKind.MM_APP_FAILED,
            entity=alert_objects.EventEntities(
                kind=alert_objects.EventEntityKind.MODEL_MONITORING_APPLICATION,
                project=self.project,
                ids=[f"{self.project}_{event.body.application_name}"],
            ),
            value_dict=error_data,
        )

        mlrun.get_run_db().generate_event(
            name=alert_objects.EventKind.MM_APP_FAILED, event_data=event_data
        )
        logger.info("Event generated successfully")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import typing
from datetime import datetime

import mlrun.common
import mlrun.common.schemas.model_monitoring.constants as mm_constants

FeatureStats = typing.NewType("FeatureStats", dict[str, dict[str, typing.Any]])
Histogram = typing.NewType("Histogram", list[list])
BinCounts = typing.NewType("BinCounts", list[int])
BinEdges = typing.NewType("BinEdges", list[float])

_MAX_FLOAT = sys.float_info.max
logger = mlrun.utils.create_logger(level="info", name="mm_helpers")


def parse_model_endpoint_project_prefix(path: str, project_name: str):
    return path.split(project_name, 1)[0] + project_name


def parse_model_endpoint_store_prefix(store_prefix: str):
    endpoint, parsed_url = mlrun.platforms.iguazio.parse_path(store_prefix)
    container, path = parsed_url.split("/", 1)
    return endpoint, container, path


def get_kafka_topic(project: str, function_name: typing.Optional[str] = None) -> str:
    if (
        function_name is None
        or function_name == mm_constants.MonitoringFunctionNames.STREAM
    ):
        function_specifier = ""
    else:
        function_specifier = f"_{function_name}"

    return (
        f"monitoring_stream_{mlrun.mlconf.system_id}_{project}{function_specifier}_v1"
    )


def _get_counts(hist: Histogram) -> BinCounts:
    """Return the histogram counts"""
    return BinCounts(hist[0])


def _get_edges(hist: Histogram) -> BinEdges:
    """Return the histogram edges"""
    return BinEdges(hist[1])


def pad_hist(hist: Histogram) -> None:
    """
    Add [-inf, x_0] and [x_n, inf] bins to the histogram inplace unless present
    """
    counts = _get_counts(hist)
    edges = _get_edges(hist)

    is_padded = edges[0] == -_MAX_FLOAT and edges[-1] == _MAX_FLOAT
    if is_padded:
        return

    counts.insert(0, 0)
    edges.insert(0, -_MAX_FLOAT)

    counts.append(0)
    edges.append(_MAX_FLOAT)


def pad_features_hist(feature_stats: FeatureStats) -> None:
    """
    Given a feature statistics dictionary, pad the histograms with edges bins
    inplace to cover input statistics from -inf to inf.
    """
    hist_key = "hist"
    for feature in feature_stats.values():
        if hist_key in feature:
            pad_hist(Histogram(feature[hist_key]))


def get_model_endpoints_creation_task_status(
    server,
) -> tuple[
    mlrun.common.schemas.BackgroundTaskState,
    typing.Optional[datetime],
    typing.Optional[set[str]],
]:
    background_task = None
    background_task_state = mlrun.common.schemas.BackgroundTaskState.running
    background_task_check_timestamp = None
    model_endpoint_uids = None
    try:
        background_task = mlrun.get_run_db().get_project_background_task(
            server.project, server.model_endpoint_creation_task_name
        )
        background_task_check_timestamp = mlrun.utils.now_date()
        log_background_task_state(
            server, background_task.status.state, background_task_check_timestamp
        )
        background_task_state = background_task.status.state
    except mlrun.errors.MLRunNotFoundError:
        logger.warning(
            "Model endpoint creation task not found listing model endpoints",
            project=server.project,
            task_name=server.model_endpoint_creation_task_name,
        )
    if background_task is None:
        model_endpoints = mlrun.get_run_db().list_model_endpoints(
            project=server.project,
            function_name=server.function_name,
            function_tag=server.function_tag,
            tsdb_metrics=False,
        )
        if model_endpoints:
            model_endpoint_uids = {
                endpoint.metadata.uid for endpoint in model_endpoints.endpoints
            }
            logger.info(
                "Model endpoints found after background task not found, model monitoring will monitor "
                "events",
                project=server.project,
                function_name=server.function_name,
                function_tag=server.function_tag,
                uids=model_endpoint_uids,
            )
            background_task_state = mlrun.common.schemas.BackgroundTaskState.succeeded
        else:
            logger.warning(
                "Model endpoints not found after background task not found, model monitoring will not "
                "monitor events",
                project=server.project,
                function_name=server.function_name,
                function_tag=server.function_tag,
            )
            background_task_state = mlrun.common.schemas.BackgroundTaskState.failed
    return background_task_state, background_task_check_timestamp, model_endpoint_uids


def log_background_task_state(
    server,
    background_task_state: mlrun.common.schemas.BackgroundTaskState,
    background_task_check_timestamp: typing.Optional[datetime],
):
    logger.info(
        "Checking model endpoint creation task status",
        task_name=server.model_endpoint_creation_task_name,
    )
    if (
        background_task_state
        in mlrun.common.schemas.BackgroundTaskState.terminal_states()
    ):
        logger.info(
            f"Model endpoint creation task completed with state {background_task_state}"
        )
    else:  # in progress
        logger.info(
            f"Model endpoint creation task is still in progress with the current state: "
            f"{background_task_state}. Events will not be monitored for the next "
            f"{mlrun.mlconf.model_endpoint_monitoring.model_endpoint_creation_check_period} seconds",
            function_name=server.function_name,
            background_task_check_timestamp=background_task_check_timestamp.isoformat(),
        )



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from collections.abc import Iterable
from typing import Optional, Union

from mlrun.artifacts import DocumentArtifact


def find_existing_attribute(obj, base_name="name", parent_name="collection"):
    # Define all possible patterns

    return None


def _extract_collection_name(vectorstore: "VectorStore") -> str:  # noqa: F821
    patterns = [
        "collection.name",
        "collection._name",
        "_collection.name",
        "_collection._name",
        "collection_name",
        "_collection_name",
    ]

    def resolve_attribute(obj, pattern):
        if "." in pattern:
            parts = pattern.split(".")
            current = vectorstore
            for part in parts:
                if hasattr(current, part):
                    current = getattr(current, part)
                else:
                    return None
            return current
        else:
            return getattr(obj, pattern, None)

    if type(vectorstore).__name__ == "PineconeVectorStore":
        try:
            url = (
                vectorstore._index.config.host
                if hasattr(vectorstore._index, "config")
                else vectorstore._index._config.host
            )
            index_name = url.split("//")[1].split("-")[0]
            return index_name
        except Exception:
            pass

    for pattern in patterns:
        try:
            value = resolve_attribute(vectorstore, pattern)
            if value is not None:
                return value
        except (AttributeError, TypeError):
            continue

    # If we get here, we couldn't find a valid collection name
    raise ValueError(
        "Failed to extract collection name from the vector store. "
        "Please provide the collection name explicitly. "
    )


class VectorStoreCollection:
    """
    A wrapper class for vector store collections with MLRun integration.

    This class wraps a vector store implementation (like Milvus, Chroma) and provides
    integration with MLRun context for document and artifact management. It delegates
    most operations to the underlying vector store while handling MLRun-specific
    functionality.

    The class implements attribute delegation through __getattr__ and __setattr__,
    allowing direct access to the underlying vector store's methods and attributes
    while maintaining MLRun integration.
    """

    def __init__(
        self,
        mlrun_context: Union["MlrunProject", "MLClientCtx"],  # noqa: F821
        vector_store: "VectorStore",  # noqa: F821
        collection_name: Optional[str] = None,
    ):
        self._collection_impl = vector_store
        self._mlrun_context = mlrun_context
        self.collection_name = collection_name or _extract_collection_name(vector_store)

    @property
    def __class__(self):
        # Make isinstance() check the wrapped object's class
        return self._collection_impl.__class__

    def __getattr__(self, name):
        # This method is called when an attribute is not found in the usual places
        # Forward the attribute access to _collection_impl
        return getattr(self._collection_impl, name)

    def __setattr__(self, name, value):
        if name in ["_collection_impl", "_mlrun_context"] or name in self.__dict__:
            # Use the base class method to avoid recursion
            super().__setattr__(name, value)
        else:
            # Forward the attribute setting to _collection_impl
            setattr(self._collection_impl, name, value)

    def _get_mlrun_project_name(self):
        import mlrun

        if self._mlrun_context and isinstance(
            self._mlrun_context, mlrun.projects.MlrunProject
        ):
            return self._mlrun_context.name
        if self._mlrun_context and isinstance(
            self._mlrun_context, mlrun.execution.MLClientCtx
        ):
            return self._mlrun_context.get_project_object().name
        return None

    def delete(self, *args, **kwargs):
        self._collection_impl.delete(*args, **kwargs)

    def add_documents(
        self,
        documents: list["Document"],  # noqa: F821
        **kwargs,
    ):
        """
        Add a list of documents to the collection.

        If the instance has an MLRun context, it will update the MLRun artifacts
        associated with the documents.

        Args:
            documents (list[Document]): A list of Document objects to be added.
            **kwargs: Additional keyword arguments to be passed to the underlying
                      collection implementation.

        Returns:
            The result of the underlying collection implementation's add_documents method.
        """
        if self._mlrun_context:
            for document in documents:
                mlrun_key = document.metadata.get(
                    DocumentArtifact.METADATA_ARTIFACT_KEY, None
                )
                mlrun_project = document.metadata.get(
                    DocumentArtifact.METADATA_ARTIFACT_PROJECT, None
                )

                if mlrun_key and mlrun_project == self._get_mlrun_project_name():
                    mlrun_tag = document.metadata.get(
                        DocumentArtifact.METADATA_ARTIFACT_TAG, None
                    )
                    artifact = self._mlrun_context.get_artifact(
                        key=mlrun_key, tag=mlrun_tag
                    )
                    if artifact.collection_add(self.collection_name):
                        self._mlrun_context.update_artifact(artifact)

        return self._collection_impl.add_documents(documents, **kwargs)

    def add_artifacts(self, artifacts: list[DocumentArtifact], splitter=None, **kwargs):
        """
        Add a list of DocumentArtifact objects to the vector store collection.

        Converts artifacts to LangChain documents, adds them to the vector store, and
        updates the MLRun context. If documents are split, the IDs are handled appropriately.

        :param artifacts: List of DocumentArtifact objects to add
        :type artifacts: list[DocumentArtifact]
        :param splitter: Document splitter to break artifacts into smaller chunks.
                        If None, each artifact becomes a single document.
        :type splitter: TextSplitter, optional
        :param kwargs: Additional arguments passed to the underlying add_documents method.
                    Special handling for 'ids' kwarg:

                    * If provided and document is split, IDs are generated as "{original_id}_{i}"
                        where i starts from 1 (e.g., "doc1_1", "doc1_2", etc.)
                    * If provided and document isn't split, original IDs are used as-is

        :return: List of IDs for all added documents. When no custom IDs are provided:

                * Without splitting: Vector store generates IDs automatically
                * With splitting: Vector store generates separate IDs for each chunk

                When custom IDs are provided:

                * Without splitting: Uses provided IDs directly
                * With splitting: Generates sequential IDs as "{original_id}_{i}" for each chunk
        :rtype: list

        """
        all_ids = []
        user_ids = kwargs.pop("ids", None)

        if user_ids:
            if not isinstance(user_ids, Iterable):
                raise ValueError("IDs must be an iterable collection")
            if len(user_ids) != len(artifacts):
                raise ValueError(
                    "The number of IDs should match the number of artifacts"
                )
        for index, artifact in enumerate(artifacts):
            documents = artifact.to_langchain_documents(splitter)
            if artifact.collection_add(self.collection_name) and self._mlrun_context:
                self._mlrun_context.update_artifact(artifact)
            if user_ids:
                num_of_documents = len(documents)
                if num_of_documents > 1:
                    ids_to_pass = [
                        f"{user_ids[index]}_{i}" for i in range(1, num_of_documents + 1)
                    ]
                else:
                    ids_to_pass = [user_ids[index]]
                kwargs["ids"] = ids_to_pass
            ids = self._collection_impl.add_documents(documents, **kwargs)
            all_ids.extend(ids)
        return all_ids

    def remove_from_artifact(self, artifact: DocumentArtifact):
        """
        Remove the current object from the given artifact's collection and update the artifact.

        Args:
            artifact (DocumentArtifact): The artifact from which the current object should be removed.
        """

        if artifact.collection_remove(self.collection_name) and self._mlrun_context:
            self._mlrun_context.update_artifact(artifact)

    def delete_artifacts(self, artifacts: list[DocumentArtifact]):
        """
        Delete a list of DocumentArtifact objects from the collection.

        This method removes the specified artifacts from the collection and updates the MLRun context.
        The deletion process varies depending on the type of the underlying collection implementation.

        Args:
            artifacts (list[DocumentArtifact]): A list of DocumentArtifact objects to be deleted.

        Raises:
            NotImplementedError: If the delete operation is not supported for the collection implementation.
        """
        store_class = self._collection_impl.__class__.__name__.lower()
        for artifact in artifacts:
            if artifact.collection_remove(self.collection_name) and self._mlrun_context:
                self._mlrun_context.update_artifact(artifact)

            if store_class == "milvus":
                expr = f"{DocumentArtifact.METADATA_SOURCE_KEY} == '{artifact.get_source()}'"
                self._collection_impl.delete(expr=expr)
            elif store_class == "chroma":
                where = {DocumentArtifact.METADATA_SOURCE_KEY: artifact.get_source()}
                self._collection_impl.delete(where=where)
            elif store_class == "pineconevectorstore":
                filter = {
                    DocumentArtifact.METADATA_SOURCE_KEY: {"$eq": artifact.get_source()}
                }
                self._collection_impl.delete(filter=filter)
            elif store_class == "mongodbatlasvectorsearch":
                filter = {DocumentArtifact.METADATA_SOURCE_KEY: artifact.get_source()}
                self._collection_impl.collection.delete_many(filter=filter)
            elif (
                hasattr(self._collection_impl, "delete")
                and "filter"
                in inspect.signature(self._collection_impl.delete).parameters
            ):
                filter = {
                    "metadata": {
                        DocumentArtifact.METADATA_SOURCE_KEY: artifact.get_source()
                    }
                }
                self._collection_impl.delete(filter=filter)
            else:
                raise NotImplementedError(
                    f"delete_artifacts() operation not supported for {store_class}"
                )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from v3io.dataplane import Client as V3IOClient
from v3io_frames import Client as V3IOFramesClient
from v3io_frames.client import ClientBase

_v3io_clients: dict[frozenset, V3IOClient] = {}
_frames_clients: dict[frozenset, ClientBase] = {}


def get_frames_client(**kwargs) -> ClientBase:
    global _frames_clients
    kw_set = frozenset(kwargs.items())
    if kw_set not in _frames_clients:
        _frames_clients[kw_set] = V3IOFramesClient(**kwargs)

    return _frames_clients[kw_set]


def get_v3io_client(**kwargs) -> V3IOClient:
    global _v3io_clients
    kw_set = frozenset(kwargs.items())
    if kw_set not in _v3io_clients:
        _v3io_clients[kw_set] = V3IOClient(**kwargs)

    return _v3io_clients[kw_set]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from os.path import expanduser

from ..config import config as mlconf
from .helpers import logger


class AzureVaultStore:
    def __init__(self, vault_name):
        try:
            from azure.identity import EnvironmentCredential
            from azure.keyvault.secrets import SecretClient
        except ImportError as exc:
            raise ImportError(
                "Azure key-vault libraries not installed, run pip install mlrun[azure-key-vault]"
            ) from exc

        self._vault_name = vault_name
        self._url = mlconf.secret_stores.azure_vault.url.format(name=vault_name)
        self._client = None

        tenant_id = self._get_secret_file_contents("tenant_id")
        client_id = self._get_secret_file_contents("client_id")

        if not tenant_id or not client_id:
            logger.info(
                "both tenant_id and client_id must be configured to use Azure vault"
            )
            return

        client_secret = self._get_secret_file_contents("secret")
        if client_secret:
            os.environ["AZURE_CLIENT_SECRET"] = client_secret

        if "AZURE_CLIENT_SECRET" not in os.environ:
            logger.info("Azure client secret could not be found")
            return

        # Azure EnvironmentCredential uses these environment variables. Populate them
        os.environ["AZURE_TENANT_ID"] = tenant_id
        os.environ["AZURE_CLIENT_ID"] = client_id

        credential = EnvironmentCredential()
        self._client = SecretClient(vault_url=self._url, credential=credential)

    @staticmethod
    def _get_secret_file_contents(file_name):
        full_path = expanduser(
            mlconf.secret_stores.azure_vault.secret_path + "/" + file_name
        )
        if os.path.isfile(full_path):
            with open(full_path) as secret_file:
                contents = secret_file.read()
            return contents
        return None

    def get_secrets(self, keys):
        # We're not checking this import, since azure-core is automatically installed by other
        # libs. Assuming we passed the checks on __init__, this is expected to work.
        from azure.core.exceptions import HttpResponseError, ResourceNotFoundError

        secrets = {}
        if not self._client:
            return secrets

        if len(keys) == 0:
            secret_details = self._client.list_properties_of_secrets()
            keys = [secret.name for secret in secret_details if secret.enabled]

        for secret in keys:
            try:
                secrets[secret] = self._client.get_secret(secret).value
            except ResourceNotFoundError:
                logger.warning(f"Secret '{secret}' is not available in Azure key vault")
            except HttpResponseError as exc:
                logger.warning(f"Exception retrieving secret '{secret}': {exc.error}")
        return secrets



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
import inspect
from abc import ABC
from types import FunctionType, MethodType
from typing import Any, Generic, Optional, Union

from .utils import CommonTypes


class MLRunInterface(ABC, Generic[CommonTypes.MLRunInterfaceableType]):
    """
    An abstract class for enriching an object interface with the properties, methods and functions written below.

    A class inheriting MLRun interface should insert what ever it needs to be inserted to the object to the following
    attributes: '_PROPERTIES', '_METHODS' and '_FUNCTIONS'. Then it should implement 'add_interface' and call 'super'.

    In order to replace object's attributes, the attributes to replace are needed to be added to the attributes:
    '_REPLACED_PROPERTIES', '_REPLACED_METHODS' and '_REPLACED_FUNCTIONS'. The original attribute will be kept in a
    backup attribute with the prefix noted in '_ORIGINAL_ATTRIBUTE_NAME'. The replacing functions / methods will be
    located by looking for the prefix noted in '_REPLACING_ATTRIBUTE_NAME'. The replacing function / method can be an
    MLRunInterface class method that return a function / method.

    For example: if "x" is in the list then the method "object.x" will be stored as "object.original_x" and "object.x"
    will then point to the method "MLRunInterface.mlrun_x". If "mlrun_x" is a class method, it will point to the
    function returned from MLRunInterface.mlrun_x()".
    """

    # Attributes to be inserted so the MLRun interface will be fully enabled.
    _PROPERTIES = {}  # type: Dict[str, Any]
    _METHODS = []  # type: List[str]
    _FUNCTIONS = []  # type: List[str]

    # Attributes to replace so the MLRun interface will be fully enabled.
    _REPLACED_PROPERTIES = {}  # type: Dict[str, Any]
    _REPLACED_METHODS = []  # type: List[str]
    _REPLACED_FUNCTIONS = []  # type: List[str]

    # TODO: Add _OPTIONALLY_REPLACED_PROPERTIES, _OPTIONALLY_REPLACED_METHODS and _OPTIONALLY_REPLACED_FUNCTIONS

    # Name template for the replaced attribute to be stored as in the object.
    _ORIGINAL_ATTRIBUTE_NAME = "original_{}"

    # Name template of the function / method to look for to replace the original function / method.
    _REPLACING_ATTRIBUTE_NAME = "mlrun_{}"

    @classmethod
    def add_interface(
        cls,
        obj: CommonTypes.MLRunInterfaceableType,
        restoration: CommonTypes.MLRunInterfaceRestorationType = None,
    ):
        """
        Enrich the object with this interface properties, methods and functions so it will have this framework MLRun's
        features.

        :param obj:         The object to enrich his interface.
        :param restoration: Restoration information tuple as returned from 'remove_interface' in order to add the
                            interface in a certain state.
        """
        # Set default value to the restoration data:
        if restoration is None:
            restoration = (None, None, None)

        # Add the MLRun properties:
        cls._insert_properties(
            obj=obj,
            properties=restoration[0],
        )

        # Replace the object's properties in MLRun's properties:
        cls._replace_properties(obj=obj, properties=restoration[1])

        # Add the MLRun functions:
        cls._insert_functions(obj=obj)

        # Replace the object's functions / methods in MLRun's functions / methods:
        cls._replace_functions(obj=obj, functions=restoration[2])

    @classmethod
    def remove_interface(
        cls, obj: CommonTypes.MLRunInterfaceableType
    ) -> CommonTypes.MLRunInterfaceRestorationType:
        """
        Remove the MLRun features from the given object. The properties and replaced attributes found in the object will
        be returned.

        :param obj: The object to remove the interface from.

        :return: A tuple of interface restoration information:
                 [0] = The interface properties.
                 [1] = The replaced properties.
                 [2] = The replaced methods and functions.
        """
        # Get the interface properties from the element:
        properties = {
            attribute: getattr(obj, attribute)
            for attribute in cls._PROPERTIES
            if hasattr(obj, attribute)  # Later it will be asserted.
        }

        # Get the replaced properties from the object:
        replaced_properties = {
            attribute: getattr(obj, attribute)
            for attribute in cls._REPLACED_PROPERTIES
            if hasattr(obj, cls._ORIGINAL_ATTRIBUTE_NAME.format(attribute))
        }

        # Get the replaced methods and functions from the object:
        replaced_functions = [
            function_name
            for function_name in [*cls._REPLACED_METHODS, *cls._REPLACED_FUNCTIONS]
            if hasattr(obj, cls._ORIGINAL_ATTRIBUTE_NAME.format(function_name))
        ]

        # Restore the replaced attributes:
        for attribute_name in [
            *replaced_properties,
            *replaced_functions,
        ]:
            cls._restore_attribute(obj=obj, attribute_name=attribute_name)

        # Remove the interface from the object:
        for attribute_name in [*cls._PROPERTIES, *cls._METHODS, *cls._FUNCTIONS]:
            assert hasattr(
                obj, attribute_name
            ), f"Can't remove the attribute '{attribute_name}' as the object doesn't has it."
            # Mark it first as None so the actual object won't be deleted:
            setattr(obj, attribute_name, None)
            delattr(obj, attribute_name)

        return properties, replaced_properties, replaced_functions

    @classmethod
    def is_applied(cls, obj: CommonTypes.MLRunInterfaceableType) -> bool:
        """
        Check if the given object has MLRun interface attributes in it. Interface is applied if all of its attributes
        are found in the object. If only replaced attributes are configured in the interface, then the interface is
        applied if some of at least one is found in the object.

        :param obj: The object to check.

        :return: True if the MLRun interface is applied on the object and False if not.
        """
        # Check for the attributes:
        attributes = [*cls._PROPERTIES, *cls._METHODS, *cls._FUNCTIONS]
        if attributes:
            return all(hasattr(obj, attribute) for attribute in attributes)

        # The interface has only replaced attributes, check if at least one is in the object:
        replaced_attributes = [
            *cls._REPLACED_PROPERTIES,
            *cls._REPLACED_METHODS,
            *cls._REPLACED_FUNCTIONS,
        ]
        return any(hasattr(obj, attribute) for attribute in replaced_attributes)

    @classmethod
    def _insert_properties(
        cls,
        obj: CommonTypes.MLRunInterfaceableType,
        properties: Optional[dict[str, Any]] = None,
    ):
        """
        Insert the properties of the interface to the object. The properties default values are being copied (not deep
        copied) into the object.

        :param obj:        The object to enrich.
        :param properties: Properties to set in the object.
        """
        # Set default dictionary if there is no properties to restore:
        if properties is None:
            properties = {}

        # Verify the provided properties are supported by this interface (noted in the interface '_PROPERTIES'):
        error_properties = [
            property_name
            for property_name in properties
            if property_name not in cls._PROPERTIES
        ]
        assert not error_properties, (
            f"The following properties provided to insert to the object are not supported by this interface: "
            f"{error_properties}"
        )

        # Insert the properties, copy only default values:
        for property_name, default_value in cls._PROPERTIES.items():
            # Verify there is no property with the same name in the object:
            assert not hasattr(obj, property_name), (
                f"Can't insert the property '{property_name}' as the object already have an attribute with the same "
                f"name."
            )
            # Insert the property to the object prioritizing given values over default ones:
            value = (
                properties[property_name]
                if property_name in properties
                else copy.copy(default_value)
            )
            setattr(obj, property_name, value)

    @classmethod
    def _insert_functions(cls, obj: CommonTypes.MLRunInterfaceableType):
        """
        Insert the functions / methods of the interface to the object.

        :param obj: The object to enrich.
        """
        # Insert the functions / methods:
        for function_name in [*cls._METHODS, *cls._FUNCTIONS]:
            # Verify there is no function / method with the same name in the object:
            assert not hasattr(obj, function_name), (
                f"Can't insert the function / method '{function_name}' as the object already have a function / method "
                f"with the same name. To replace a function / method, add the name of the function / method to the "
                f"'_REPLACED_METHODS' / '_REPLACED_METHODS' list and follow the instructions documented."
            )
            # Get the function / method:
            func = getattr(cls, function_name)
            # If the function is a method and not a function (appears in '_METHODS' and not '_FUNCTIONS'), set the
            # 'self' to the object:
            if function_name in cls._METHODS:
                func = MethodType(func, obj)
            # Insert the function / method to the object:
            setattr(obj, function_name, func)

    @classmethod
    def _replace_properties(
        cls,
        obj: CommonTypes.MLRunInterfaceableType,
        properties: Optional[dict[str, Any]] = None,
    ):
        """
        Replace the properties of the given object according to the configuration in the MLRun interface.

        :param obj:        The object to replace its properties.
        :param properties: The properties to replace in the object. Default: all the properties in the interface
                           '_REPLACE_PROPERTIES' dictionary.
        """
        # Set default replacing properties if there are no properties given:
        if properties is None:
            properties = {
                property_name: copy.copy(property_value)
                for property_name, property_value in cls._REPLACED_PROPERTIES.items()
            }
        else:
            # Verify the provided properties are supported by this interface (noted in the interface's
            # '_REPLACED_PROPERTIES' dictionary):
            error_properties = [
                property_name
                for property_name in properties
                if property_name not in cls._REPLACED_PROPERTIES
            ]
            assert not error_properties, (
                f"The following properties provided to be replace in the object are not supported by this "
                f"interface: {error_properties}"
            )

        # Replace the properties in the object:
        for property_name, property_value in properties.items():
            # Verify there is a property with this name in the object to replace:
            assert hasattr(
                obj, property_name
            ), f"Can't replace the property '{property_name}' as the object doesn't have a property with this name."
            # Replace the property:
            cls._replace_property(
                obj=obj,
                property_name=property_name,
                property_value=property_value,
                include_none=True,
            )

    @classmethod
    def _replace_functions(
        cls,
        obj: CommonTypes.MLRunInterfaceableType,
        functions: Optional[list[str]] = None,
    ):
        """
        Replace the functions / methods of the given object according to the configuration in the MLRun interface.

        :param obj:       The object to replace its functions / methods.
        :param functions: The functions / methods to replace in the object. Default: all the functions / methods in
                          the interface '_REPLACE_METHODS' and '_REPLACE_FUNCTIONS' lists.
        """
        # Set default list if there are no functions / methods to restore:
        if functions is None:
            functions = [*cls._REPLACED_METHODS, *cls._REPLACED_FUNCTIONS]
        else:
            # Verify the provided functions / methods are supported by this interface (noted in the interface's
            # '_REPLACED_METHODS' or '_REPLACE_FUNCTIONS' lists):
            error_functions = [
                function_name
                for function_name in functions
                if function_name
                not in [*cls._REPLACED_METHODS, *cls._REPLACED_FUNCTIONS]
            ]
            assert not error_functions, (
                f"The following functions / methods provided to be replace in the object are not supported by this "
                f"interface: {error_functions}"
            )

        # Replace the functions / methods in the object:
        for function_name in functions:
            # Verify there is a function / method with this name in the object to replace:
            assert hasattr(obj, function_name), (
                f"Can't replace the function / method '{function_name}' as the object doesn't have a function / method "
                f"with this name."
            )
            # Replace the method:
            cls._replace_function(obj=obj, function_name=function_name)

    @classmethod
    def _replace_property(
        cls,
        obj: CommonTypes.MLRunInterfaceableType,
        property_name: str,
        property_value: Any = None,
        include_none: bool = False,
    ):
        """
        Replace the property in the object with the configured property in this interface. The original property will be
        stored in a backup attribute with the prefix noted in '_ORIGINAL_ATTRIBUTE_NAME' and the replacing property
        will be the one with the prefix noted in '_REPLACING_ATTRIBUTE_NAME'. If the property value should be None, set
        'include_none' to True, otherwise the interface default will be copied if 'property_value' is None.

        :param obj:            The object to replace its property.
        :param property_name:  The property name to replace.
        :param property_value: The value to set. If not provided, the interface's default will be copied.
        :param include_none:   Whether to enable the property value to be None.
        """
        # Get the original property from the object:
        original_property = getattr(obj, property_name)

        # Set a backup attribute with for the original property:
        original_property_name = cls._ORIGINAL_ATTRIBUTE_NAME.format(property_name)
        setattr(obj, original_property_name, original_property)

        # Check if a value is provided, if not copy the default value in this interface if None should not be included:
        if not include_none and property_value is None:
            property_value = copy.copy(cls._REPLACED_PROPERTIES[property_name])

        # Replace the property:
        setattr(obj, property_name, property_value)

    @classmethod
    def _replace_function(
        cls, obj: CommonTypes.MLRunInterfaceableType, function_name: str
    ):
        """
        Replace the method / function in the object with the configured method / function in this interface. The
        original method / function will be stored in a backup attribute with the prefix noted in
        '_ORIGINAL_ATTRIBUTE_NAME' and the replacing method / function will be the one with the prefix noted in
        '_REPLACING_ATTRIBUTE_NAME'.

        :param obj:           The object to replace its method.
        :param function_name: The method / function name to replace.
        """
        # Get the original function from the object:
        original_function = getattr(obj, function_name)

        # Set a backup attribute with for the original function:
        original_function_name = cls._ORIGINAL_ATTRIBUTE_NAME.format(function_name)
        setattr(obj, original_function_name, original_function)

        # Get the function to replace from the interface:
        replacing_function_name = cls._REPLACING_ATTRIBUTE_NAME.format(function_name)
        replacing_function = getattr(cls, replacing_function_name)

        # Check if the replacing function is a class method (returning a function to use as the replacing function):
        if inspect.ismethod(replacing_function):
            replacing_function = replacing_function()

        # Wrap the replacing function with 'functools.wraps' decorator so the properties of the original function will
        # be passed to the replacing function:
        replacing_function = functools.wraps(original_function)(replacing_function)

        # If the replacing function is a method and not a function (appears in the _REPLACED_METHODS and not
        # _REPLACED_FUNCTIONS), set the 'self' to the object:
        if function_name in cls._REPLACED_METHODS:
            replacing_function = MethodType(replacing_function, obj)

        # Replace the function:
        setattr(obj, function_name, replacing_function)

    @classmethod
    def _restore_attribute(
        cls, obj: CommonTypes.MLRunInterfaceableType, attribute_name: str
    ):
        """
        Restore the replaced attribute (property, method or function) in the object, removing the backup attribute as
        well.

        :param obj:            The object to restore its method.
        :param attribute_name: The method to restore.
        """
        # Get the original attribute:
        original_attribute_name = cls._ORIGINAL_ATTRIBUTE_NAME.format(attribute_name)
        original_attribute = getattr(obj, original_attribute_name)

        # Set the attribute to point back to the original attribute:
        setattr(obj, attribute_name, original_attribute)

        # Remove the original backup attribute:
        setattr(obj, original_attribute_name, None)
        delattr(obj, original_attribute_name)

    @staticmethod
    def _get_function_argument(
        func: FunctionType,
        argument_name: str,
        passed_args: Optional[tuple] = None,
        passed_kwargs: Optional[dict] = None,
        default_value: Any = None,
    ) -> tuple[Any, Union[str, int, None]]:
        """
        Get a passed argument (from *args or **kwargs) to a function. If the argument was not found the default value
        will be returned. In addition, the keyword of the argument in `kwargs` or the index of the argument in `args`
        will be returned as well.

        :param func:          The function that is being called.
        :param argument_name: The argument name to get.
        :param passed_args:   The passed arguments to the function (*args).
        :param passed_kwargs: The passed keyword arguments to the function (*kwargs).
        :param default_value: The default value to use in case it was not passed.

        :return: A tuple of:
                 [0] = The argument value or the default value if it was not found in any of the arguments.
                 [1] = If it was found in `kwargs` - the keyword of the argument. If it was found in `args` - the index
                       of the argument. If it was not found, None.
        """
        # Set default values for arguments data structures:
        if passed_args is None:
            passed_args = []
        if passed_kwargs is None:
            passed_kwargs = {}

        # Check in the key word arguments first:
        if argument_name in passed_kwargs:
            return passed_kwargs[argument_name], argument_name

        # Check in the arguments, inspecting the function's parameters to get the right index:
        func_parameters = {
            parameter_name: i
            for i, parameter_name in enumerate(
                inspect.signature(func).parameters.keys()
            )
        }
        if (
            argument_name in func_parameters
            and len(passed_args) >= func_parameters[argument_name] + 1
        ):
            return (
                passed_args[func_parameters[argument_name]],
                func_parameters[argument_name],
            )

        # The argument name was not found:
        return default_value, None
